# SPDX-License-Identifier: GPL-2.0

import sys

def will_be_deprecated(feature, deadline, additional_notice=''):
    sys.stderr.write('\n'.join([
'',
'WARNING: %s will be deprecated by %s.' % (feature, deadline),
'    %s' % additional_notice,
'    Please report your usecase to Github issues[1], sj@kernel.org,',
'    damon@lists.linux.dev and/or linux-mm@kvack.org if you depend on those.',
'',
'    [1] https://github.com/damonitor/damo/issues',
'',
'']))

def deprecated(feature, deadline, do_exit=False, exit_code=1,
        additional_notice=''):
    sys.stderr.write('\n'.join([
'',
'WARNING: %s is deprecated.' % feature,
'    The support will be removed by %s.' % deadline,
'    %s' % additional_notice,
'    Please report your usecase to Github issues[1], sj@kernel.org,',
'    damon@lists.linux.dev and/or linux-mm@kvack.org if you depend on those.',
'',
'    [1] https://github.com/damonitor/damo/issues',
'',
'']))
    if do_exit:
        exit(exit_code)



# SPDX-License-Identifier: GPL-2.0

import os
import random
import time

import _damo_fmt_str
import _damo_records
import damo_record_info

page_map = {}

sz_page = 4096

def get_page(pfn):
    if not pfn in page_map:
        page_map[pfn] = bytearray(sz_page)
    return page_map[pfn]

def access_region(start_addr, end_addr):
    for addr in range(start_addr, end_addr, sz_page):
        page = get_page(addr / sz_page)
        just_for_access = page[int(sz_page / 2)]

def replay_snapshot(snapshot, mon_intervals):
    runtime_sec = (snapshot.end_time - snapshot.start_time) / 1000000000
    max_nr_accesses = mon_intervals.aggr / mon_intervals.sample
    time_slice = runtime_sec / max_nr_accesses
    nr_slices = int(runtime_sec / time_slice)
    for slice_idx in range(nr_slices):
        start_time = time.time()
        for region in snapshot.regions:
            if slice_idx < region.nr_accesses.samples:
                access_region(region.start, region.end)
        while time.time() - start_time < time_slice:
            pass

def test_perf(size_mem, test_time):
    test_start_time = time.time()
    access_start_time = time.time()
    next_output_time = time.time() + 1
    nr_accesses = 0
    while True:
        access_region(0, size_mem)
        nr_accesses += 1

        if time.time() >= next_output_time:
            print('replayer can access %s memory per second' %
                  _damo_fmt_str.format_sz(
                      size_mem * nr_accesses /
                      (time.time() - access_start_time),
                      machine_friendly=False))
            next_output_time = time.time() + 1
            nr_accesses = 0
            access_start_time = time.time()

        if time.time() - test_start_time >= test_time:
            break

def main(args):
    if args.test_perf is True:
        size_mem = _damo_fmt_str.text_to_bytes(args.test_perf_sz_mem)
        test_time = _damo_fmt_str.text_to_sec(args.test_perf_runtime)
        return test_perf(size_mem, test_time)

    input_file = args.input

    if not os.path.isfile(input_file):
        print('input file (%s) not exists' % input_file)
        exit(1)

    records, err = _damo_records.get_records(record_file=input_file)
    if err:
        print('parsing damon records file (%s) failed (%s)' %
              (input_file, err))
        exit(1)

    if len(records) == 0:
        print('no monitoring records in the file')
        exit(1)

    if len(records) != 1:
        print('supporting only single record for now')

    record = records[0]
    for snapshot in record.snapshots:
        for region in snapshot.regions:
            region.nr_accesses.add_unset_unit(record.intervals)

    if args.progress_notice_interval is None:
        progress_notice_interval = 3
    else:
        progress_notice_interval = _damo_fmt_str.text_to_sec(
                args.progress_notice_interval)

    play_start_time = time.time()
    progress_notice_time = play_start_time + + progress_notice_interval
    for idx, snapshot in enumerate(record.snapshots):
        replay_snapshot(snapshot, record.intervals)
        if time.time() >= progress_notice_time:
            print('%d/%d snapshots replayed in %.3f seconds' %
                  (idx, len(record.snapshots), time.time() - play_start_time))
            progress_notice_time += progress_notice_interval
    print('all snapshots are replayed')

def set_argparser(parser):
    parser.add_argument('input', metavar='<file>', nargs='?',
                        default='damon.data',
                        help='record file to replay')
    parser.add_argument('--progress_notice_interval', metavar='<seconds>',
                        help='time interval between replay progress notice')
    parser.add_argument('--test_perf', action='store_true',
                        help='measure performance of replayer')
    parser.add_argument('--test_perf_sz_mem', metavar='<bytes>',
                        default='1 GiB',
                        help='size of memory to use for --test_perf')
    parser.add_argument('--test_perf_runtime', metavar='<seconds>',
                        default='1 m',
                        help='runtime of --test_perf')
    parser.description = 'Replay monitored access pattern'
    return parser



# SPDX-License-Identifier: GPL-2.0

import argparse

import damo_report_access

def main(args):
    if args.subtopic == 'record_format_keywords':
        print('\n\n'.join(
            ['%s' % f for f in damo_report_access.record_formatters]))
        return
    if args.subtopic == 'snapshot_format_keywords':
        print('\n\n'.join(
            ['%s' % f for f in damo_report_access.snapshot_formatters]))
        return
    if args.subtopic == 'region_format_keywords':
        print('\n\n'.join(
            ['%s' % f for f in damo_report_access.region_formatters]))
        return

    print('command line options for monitoring results visualization options')
    parser = argparse.ArgumentParser(add_help=False)
    damo_report_access.add_fmt_args(parser)
    help_msg = parser.format_help()
    pars = help_msg.split('\n\n')
    # the first paragraph is usage for this virtual command.  Ignore.
    pars = pars[1:]
    if len(pars) == 1:
        print('\n'.join(pars[0].split('\n')[1:]))
    else:
        print('\n\n'.join(pars))

def set_argparser(parser):
    parser.description = 'help for DAMON monitoring results visualization options'
    parser.add_argument(
            'subtopic', nargs='?',
            default='options',
            choices=['options', 'record_format_keywords',
                     'snapshot_format_keywords', 'region_format_keywords'],
            help='sub-topic to get help')
    return parser



# SPDX-License-Identifier: GPL-2.0

"""
Record monitored data access patterns.
"""

import json
import os
import signal
import subprocess
import time

import _damo_fmt_str
import _damo_records
import _damon
import _damon_args

class DataForCleanup:
    kdamonds_idxs = None
    orig_kdamonds = None
    record_handle = None

data_for_cleanup = DataForCleanup()

cleaning = False

def cleanup_exit(exit_code):
    global cleaning
    if cleaning == True:
        return
    cleaning = True
    if data_for_cleanup.kdamonds_idxs != None:
        # ignore returning error, as kdamonds may already finished
        _damon.turn_damon_off(data_for_cleanup.kdamonds_idxs)
        err = _damon.stage_kdamonds(data_for_cleanup.orig_kdamonds)
        if err:
            print('failed restoring previous kdamonds setup (%s)' % err)

    if data_for_cleanup.record_handle:
        _damo_records.finish_recording(data_for_cleanup.record_handle)

    exit(exit_code)

def sighandler(signum, frame):
    print('\nsignal %s received' % signum)
    cleanup_exit(signum)

def handle_args(args):
    if _damon.any_kdamond_running() and not args.deducible_target:
        args.deducible_target = 'ongoing'

    args.output_permission, err = _damo_records.parse_file_permission_str(
            args.output_permission)
    if err != None:
        print('wrong --output permission (%s) (%s)' %
                (args.output_permission, err))
        exit(1)

    # backup duplicate output file
    if os.path.isfile(args.out):
        os.rename(args.out, args.out + '.old')


    if 'mem_footprint' in args.do_record:
        footprint_file_path = '%s.mem_footprint' % args.out
        if os.path.isfile(footprint_file_path):
            os.rename(footprint_file_path, footprint_file_path + '.old')

    err = _damo_records.set_perf_path(args.perf_path)
    if err != None:
        print(err)
        exit(-3)

def tracepoint_from_args(args):
    if not 'access' in args.do_record or args.snapshot is not None:
        return None
    if args.schemes_target_regions is False:
        return _damo_records.perf_event_damon_aggregated
    return _damo_records.perf_event_damos_before_apply

def snapshot_requests_from_args(args):
    ''' Returns snapshot_request and error '''
    if not 'access' in args.do_record or args.snapshot is None:
        return None, None

    tried_regions_of = None
    if args.schemes_target_regions:
        tried_regions_of = []
        for kidx, kd in enumerate(kdamonds):
            for cidx, ctx in enumerate(kd.contexts):
                for sidx, scheme in enumerate(ctx.schemes):
                    tried_regions_of.append([kidx, cidx, sidx])

    dfilters, err = _damon_args.damos_options_to_filters(
            args.snapshot_damos_filter)
    if err is not None:
        return None, 'wrong --snapshot_damos_filters (%s)' % err

    record_filter, err = _damo_records.args_to_filter(args)
    if err is not None:
        return None, err

    return _damo_records.RecordGetRequest(
            tried_regions_of=tried_regions_of, record_file=None,
            snapshot_damos_filters=dfilters,
            record_filter=record_filter, total_sz_only=False,
            dont_merge_regions=False), None

def mk_handle(args, kdamonds, monitoring_intervals):
    tracepoint = tracepoint_from_args(args)
    snapshot_request, err = snapshot_requests_from_args(args)
    if err is not None:
        print(err)
        cleanup_exit(1)
    if snapshot_request is not None:
        snapshot_interval_sec = _damo_fmt_str.text_to_sec(
                args.snapshot[0])
        snapshot_count = _damo_fmt_str.text_to_nr(args.snapshot[1])
    else:
        snapshot_interval_sec = None
        snapshot_count = None

    handle = _damo_records.RecordingHandle(
            # for access pattern monitoring
            tracepoint=tracepoint, file_path=args.out,
            file_format=args.output_type,
            file_permission=args.output_permission,
            monitoring_intervals=monitoring_intervals,
            # for perf profile
            do_profile='cpu_profile' in args.do_record,
            # for children processes recording and memory footprint
            kdamonds=kdamonds, add_child_tasks=args.include_child_tasks,
            record_mem_footprint='mem_footprint' in args.do_record,
            record_vmas='vmas' in args.do_record,
            record_proc_stats='proc_stats' in args.do_record,
            timeout=args.timeout, snapshot_request=snapshot_request,
            snapshot_interval_sec=snapshot_interval_sec,
            snapshot_count=snapshot_count)

    return handle

def main(args):
    global data_for_cleanup

    _damon.ensure_root_and_initialized(args)

    handle_args(args)

    # Setup for cleanup
    data_for_cleanup.orig_kdamonds = _damon.current_kdamonds()
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    # Now the real works
    if not _damon_args.is_ongoing_target(args):
        err, kdamonds = _damon_args.turn_damon_on(args)
        if err:
            print('could not turn DAMON on (%s)' % err)
            cleanup_exit(-2)
        data_for_cleanup.kdamonds_idxs = ['%d' % idx
                for idx, k in enumerate(kdamonds)]
        # TODO: Support multiple kdamonds, multiple contexts
        monitoring_intervals = kdamonds[0].contexts[0].intervals
        now_kdamonds = _damon.current_kdamonds()
        kdamonds[0].pid = now_kdamonds[0].pid
    else:
        if not _damon.any_kdamond_running():
            print('DAMON is not turned on')
            exit(1)

        # TODO: Support multiple kdamonds, multiple contexts
        monitoring_intervals = data_for_cleanup.orig_kdamonds[
                0].contexts[0].intervals
        kdamonds = data_for_cleanup.orig_kdamonds

    record_handle = mk_handle(args, kdamonds, monitoring_intervals)

    data_for_cleanup.record_handle = record_handle

    if record_handle.will_take_awhile():
        print('Press Ctrl+C to stop')
    _damo_records.start_recording(record_handle)
    cleanup_exit(0)

def set_argparser(parser):
    _damon_args.set_argparser(parser, add_record_options=True, min_help=True)
    parser.add_argument('--output_type',
                        choices=_damo_records.file_types,
                        default=_damo_records.file_type_json_compressed,
                        help='output file\'s type')
    parser.add_argument('--output_permission', type=str, default='600',
                        help='permission of the output file')
    parser.add_argument('--perf_path', type=str, default='perf',
                        help='path of perf tool ')
    parser.add_argument('--exclude_child_tasks', action='store_false',
                        dest='include_child_tasks',
                        help='do not record access of child processes')
    parser.add_argument('--include_child_tasks', action='store_true',
                        help='record accesses of child processes')
    parser.add_argument('--schemes_target_regions', action='store_true',
                        help='record schemes tried to be applied regions')
    parser.add_argument('--snapshot', metavar=('<delay>', '<count>'), nargs=2,
                        help='record accesses as snapshots')
    parser.add_argument('--timeout', type=float, metavar='<seconds>',
                        help='stop recording after the given seconds')
    parser.add_argument('--do_record', nargs='+',
                        default=['access', 'cpu_profile', 'mem_footprint',
                                 'vmas', 'proc_stats'],
                        choices=['access', 'cpu_profile', 'mem_footprint',
                                 'vmas', 'proc_stats'],
                        help='what to do record')
    _damo_records.set_snapshot_damos_filters_option(parser)
    _damo_records.set_filter_argparser(parser)
    return parser



# SPDX-License-Identifier: GPL-2.0

import argparse
import code
import copy
import json
import math
import os
import signal
import sys
import tempfile
import time

import _damo_ascii_color
import _damo_fmt_str
import _damo_print
import _damo_records
import _damon
import _damon_args

class Formatter:
    keyword = None
    format_fn = None
    help_msg = None

    def __init__(self, keyword, format_fn, help_msg):
            self.keyword = keyword
            self.format_fn = format_fn
            self.help_msg = help_msg

    def __str__(self):
        return '%s\n%s' % (self.keyword, self.help_msg)

record_formatters = [
        Formatter('<kdamond index>',
            lambda record, fmt: '%s' % record.kdamond_idx,
            'index of the record\'s kdamond'),
        Formatter('<context index>',
            lambda record, fmt: '%s' % record.context_idx,
            'index of the record\'s DAMON context'),
        Formatter('<scheme index>',
            lambda record, fmt: '%s' % record.scheme_idx,
            'index of the record\'s DAMOS scheme'),
        Formatter('<target id>',
            lambda record, fmt: '%s' % record.target_id,
            'index of the record\'s DAMON target'),
        Formatter('<abs start time>',
            lambda record, fmt:
            _damo_fmt_str.format_time_ns(record.snapshots[0].start_time,
                                         fmt.raw_number),
            'absolute time of the start of the record'),
        Formatter('<duration>',
            lambda record, fmt:
            _damo_fmt_str.format_time_ns(
                record.snapshots[-1].end_time - record.snapshots[0].start_time,
                fmt.raw_number),
            'duration of the record'),
        Formatter('<intervals>',
                  lambda record, fmt: record_intervals(record, fmt.raw_number),
                  'monitoring intervals'),
        Formatter('<intervals goal>',
                  lambda record, fmt:
                  record.intervals.intervals_goal.to_str(fmt.raw_number),
                  'monitoring intervals'),
        Formatter('<format strings>',
                  lambda record, fmt: format_strings(fmt),
                  'current format strings')
        ]

snapshot_formatters = [
        Formatter('<total bytes>',
            lambda snapshot, record, fmt:
            _damo_fmt_str.format_sz(snapshot.total_bytes, fmt.raw_number),
            'total bytes of regions in the snapshot'),
        Formatter('<duration>',
            lambda snapshot, record, fmt:
                  _damo_fmt_str.format_time_ns(
                      snapshot.end_time - snapshot.start_time, fmt.raw_number),
                  'access monitoring duration for the snapshot'),
        Formatter('<start time>',
            lambda snapshot, record, fmt:
                  _damo_fmt_str.format_time_ns(
                      snapshot.start_time -
                      record.snapshots[0].start_time, fmt.raw_number),
            'access monitoring start time for the snapshot, relative to the record start time'),
        Formatter('<end time>',
            lambda snapshot, record, fmt:
                  _damo_fmt_str.format_time_ns(
                      snapshot.end_time - record.snapshots[0].start_time,
                      fmt.raw_number),
            'access monitoring end time for the snapshot, relative to the record end time'),
        Formatter('<abs start time>',
            lambda snapshot, record, fmt:
            _damo_fmt_str.format_time_ns(snapshot.start_time, fmt.raw_number),
            'absolute access monitoring start time for the snapshot'),
        Formatter('<abs end time>',
            lambda snapshot, record, fmt:
            _damo_fmt_str.format_time_ns(snapshot.end_time, fmt.raw_number),
            'absolute access monitoring end time for the snapshot'),
        Formatter('<number of regions>',
            lambda snapshot, record, fmt:
            _damo_fmt_str.format_nr(len(snapshot.regions), fmt.raw_number),
            'the number of regions in the snapshot'),
        Formatter('<region box colors>',
            lambda snapshot, record, fmt:
            _damo_ascii_color.color_samples(fmt.region_box_format.colorset),
            'available colors for the region box'),
        Formatter('<region box description>',
            lambda snapshot, record, fmt:
            fmt.region_box_format.description_msg(fmt.raw_number),
            'description about region box (what and how it represents)'),
        Formatter('<temperature-sz histogram>',
                  lambda snapshot, record, fmt:
                  temperature_sz_hist_str(snapshot, record, fmt, False),
                  'temperature to total size of the regions histogram'),
        Formatter('<temperature-df-passed-sz histogram>',
                  lambda snapshot, record, fmt:
                  temperature_sz_hist_str(snapshot, record, fmt, True),
                  ' '.join(['temperature to total size of DAMOS filters (df)',
                            'passed regions histogram'])),
        Formatter('<recency-sz histogram>',
                  lambda snapshot, record, fmt:
                  recency_hist_str(snapshot, record, fmt, False),
                  'last accessed time to total size of the regions histogram'),
        Formatter('<recency-df-passed-sz histogram>',
                  lambda snapshot, record, fmt:
                  recency_hist_str(snapshot, record, fmt, True),
                  ' '.join(['last accessed time to total size of',
                            'DAMOS filters (df) passed regions histogram'])),
        Formatter('<heatmap>',
                  lambda snapshot, record, fmt:
                  heatmap_str(snapshot, record, fmt),
                  'heatmap of the snapshot'),
        Formatter('<filters passed heatmap>',
                  lambda snapshot, record, fmt:
                  df_passed_heatmap_str(snapshot, record, fmt),
                  'heatmap of the snapshot for filter-passed regions'),
        Formatter(
            '<filters passed type>',
            lambda snapshot, record, fmt: filters_pass_type_of(record),
            'type of <filters passed bytes> memory'),
        Formatter(
                '<filters passed bytes>',
                lambda snapshot, record, fmt:
                filters_passed_bytes(snapshot, fmt),
                'bytes of regions that passed DAMOS filters'),
        Formatter(
                '<positive access samples ratio>',
                lambda snapshot, record, fmt:
                positive_access_sample_ratio(snapshot, record, fmt),
                'positive access samples ratio'),
        Formatter(
                '<estimated memory bandwidth>',
                lambda snapshot, record, fmt:
                estimated_mem_bw(snapshot, record, fmt),
                'estimated memory bandwidth'),
        Formatter(
                '<filters passed estimated memory bandwidth>',
                lambda snapshot, record, fmt:
                estimated_mem_bw(snapshot, record, fmt,
                                 filter_passed_only=True),
                'estimated memory bandwidth'),
        Formatter(
                '<intervals tuning status>',
                lambda snapshot, record, fmt:
                intervals_tuning_status(snapshot, record, fmt),
                'intervals tuning status'),
        Formatter(
                '<damos stats>',
                lambda snapshot, record, fmt:
                damos_stats_str(snapshot, record, fmt),
                'DAMOS stats for the snapshot-retrieval scheme'),
        ]

region_formatters = [
        Formatter(
            '<index>',
            lambda index, region, fmt:
            _damo_fmt_str.format_nr(index, fmt.raw_number),
            'index of the region in the regions of the snapshot'),
        Formatter(
            '<start address>',
            lambda index, region, fmt:
            _damo_fmt_str.format_sz(region.start, fmt.raw_number),
            'start address of the region'),
        Formatter(
            '<end address>',
            lambda index, region, fmt:
            _damo_fmt_str.format_sz(region.end, fmt.raw_number),
            'end address of the region'),
        Formatter(
            '<size>',
            lambda index, region, fmt:
            _damo_fmt_str.format_sz(region.size(), fmt.raw_number),
            'size of the region'),
        Formatter(
            '<access rate>',
            lambda index, region, fmt:
            _damo_fmt_str.format_percent(region.nr_accesses.percent,
                                         fmt.raw_number),
            'monitored access rate of the region'),
        Formatter(
            '<age>',
            lambda index, region, fmt:
            _damo_fmt_str.format_time_us(region.age.usec, fmt.raw_number),
            'how long the access pattern of the region has maintained'),
        Formatter(
            '<temperature>',
            lambda index, region, fmt:
            temperature_str(region, fmt.raw_number, fmt),
            'access temperature of the region'),
        Formatter(
            '<filters passed bytes>',
            lambda index, region, fmt:
            _damo_fmt_str.format_sz(region.sz_filter_passed, fmt.raw_number)
            if region.sz_filter_passed is not None else 'N/A',
            'bytes of the region that passed DAMOS filters'),
        Formatter(
            '<box>',
            lambda index, region, fmt:
            fmt.region_box_format.to_str(region),
            'user-customizable (via --region_box_*) box (age/access_rate/size by default)'),
        ]

default_record_head_format = 'kdamond <kdamond index> / context <context index> / scheme <scheme index> / target id <target id> / recorded for <duration> from <abs start time>'
default_snapshot_head_format = 'monitored time: [<start time>, <end time>] (<duration>)\n<heatmap>'
default_snapshot_head_format_without_heatmap = 'monitored time: [<start time>, <end time>] (<duration>)'
default_region_format = '<index> addr <start address> size <size> access <access rate> age <age>'
default_snapshot_tail_format = 'memory bw estimate: <estimated memory bandwidth>\ntotal size: <total bytes>'
default_snapshot_tail_format_filter_installed = 'memory bw estimate: <estimated memory bandwidth>  df-passed: <filters passed estimated memory bandwidth>\ntotal size: <total bytes>  df-passed <filters passed bytes>'

def record_intervals(record, raw_number):
    intervals = record.intervals
    return 'sample %s, aggr %s' % (
            _damo_fmt_str.format_time_us(intervals.sample, raw_number),
            _damo_fmt_str.format_time_us(intervals.aggr, raw_number))

def format_strings(fmt):
    return '\n'.join([
        'record head: "%s"' % fmt.format_record_head,
        'snapshot head: "%s"' % fmt.format_snapshot_head,
        'region: "%s"' % fmt.format_region,
        'snapshot tail: "%s"' % fmt.format_snapshot_tail,
        'record tail: "%s"' % fmt.format_record_tail,
        ])

def filters_passed_bytes(snapshot, fmt):
    bytes = 0
    for region in snapshot.regions:
        if region.sz_filter_passed is None:
            return 'N/A'
        bytes += region.sz_filter_passed
    return _damo_fmt_str.format_sz(bytes, fmt.raw_number)

def positive_access_sample_ratio(snapshot, record, fmt):
    max_samples_per_region = record.intervals.aggr / record.intervals.sample
    max_samples = max_samples_per_region * len(snapshot.regions)
    nr_samples = 0
    for region in snapshot.regions:
        region.nr_accesses.add_unset_unit(record.intervals)
        nr_samples += region.nr_accesses.samples
    return _damo_fmt_str.format_percent(
            nr_samples * 100 / max_samples, fmt.raw_number)

def intervals_tuning_status(snapshot, record, fmt):
    max_samples_per_region = record.intervals.aggr / record.intervals.sample
    max_observation = 0
    observation = 0
    for region in snapshot.regions:
        region.nr_accesses.add_unset_unit(record.intervals)
        max_observation += max_samples_per_region * region.size()
        observation += region.size() * region.nr_accesses.samples
    return '%s (%s / %s) %s %s' % (
            _damo_fmt_str.format_percent(
                observation * 100 / max(max_observation, 1), fmt.raw_number),
            _damo_fmt_str.format_sz(observation, fmt.raw_number),
            _damo_fmt_str.format_sz(max_observation, fmt.raw_number),
            _damo_fmt_str.format_time_us(record.intervals.sample,
                                         fmt.raw_number),
            _damo_fmt_str.format_time_us(record.intervals.aggr,
                                         fmt.raw_number),
            )

def damos_stats_str(snapshot, record, fmt):
    if snapshot.damos_stats is None:
        return 'none'
    return snapshot.damos_stats.to_str(fmt.raw_number)

def estimated_mem_bw(snapshot, record, fmt, filter_passed_only=False):
    access_bytes = 0
    for region in snapshot.regions:
        if filter_passed_only is True:
            bytes = region.sz_filter_passed
        else:
            bytes = region.size()
        access_bytes += bytes * region.nr_accesses.samples
    bw_per_sec = access_bytes / (record.intervals.aggr / 1000000)
    return '%s per second' % _damo_fmt_str.format_sz(
            bw_per_sec, fmt.raw_number)

def filters_pass_type_of(record):
    ops_filters = [f for f in record.scheme_filters if f.handled_by_ops()]
    if len(ops_filters) == 0:
        return 'n/a'
    text = ', '.join(['%s' % f for f in ops_filters])
    if ops_filters[-1].allow is True:
        text += '\n# WARN: Allow filters at the end of the list means nothing on <6.15'
    return text

def get_linearscale_hist_ranges(minv, maxv, nr_ranges, cumulate):
    hist_ranges = []
    total_interval = maxv + 1 - minv
    interval = max(math.ceil(total_interval / nr_ranges), 1)
    for i in range(nr_ranges):
        if cumulate:
            start_range = minv
        else:
            start_range = minv + interval * i
        hist_ranges.append([start_range, minv + interval * (i + 1)])
    return hist_ranges

def get_logscale_hist_ranges(minv, maxv, nr_ranges, cumulate):
    initial_interval = max(int((maxv - minv) / 2**nr_ranges), 1)
    # start from 1 second interval
    ranges = [[minv, minv + initial_interval]]
    while ranges[-1][-1] < maxv:
        last_interval = ranges[-1][1] - ranges[-1][0]
        if cumulate:
            start_range = minv
        else:
            start_range = ranges[-1][-1]
        ranges.append([start_range, ranges[-1][-1] + last_interval * 2])
    return ranges

def get_hist_ranges(minv, maxv, nr_ranges, logscale, cumulate):
    if logscale:
        return get_logscale_hist_ranges(minv, maxv, nr_ranges, cumulate)
    return get_linearscale_hist_ranges(minv, maxv, nr_ranges, cumulate)

def histogram_str(hist):
    """
    Format histogram string to show.
    Args:
        hist (list): List of lines of histogram to show.
    Returns:
        str: A string of the histogram that formatted in a user-friendly way.
    """
    yvals = [entry[2] for entry in hist]
    min_yval = min(yvals)
    max_yval = max(yvals)
    max_xval_range_str = max([len(entry[0]) for entry in hist])
    max_yval_str = max([len(entry[1]) for entry in hist])
    max_dots = 20
    yval_dot_interval = max(int((max_yval - min_yval) / max_dots), 1)
    lines = []
    for xrange_str, y_str, yval in hist:
        xrange_str = '%s%s' % (xrange_str,
                               ' ' * (max_xval_range_str - len(xrange_str)))
        y_str = '%s%s' % (y_str,
                           ' ' * (max_yval_str - len(y_str)))

        nr_dots = min(math.ceil((yval - min_yval) / yval_dot_interval), max_dots)
        bar = '|%s%s|' % ('*' * nr_dots, ' ' * (max_dots - nr_dots))
        lines.append('%s %s %s' % (xrange_str, y_str, bar))
    return '\n'.join(lines)

def get_unsorted_histogram(snapshot, fmt, get_x_fn, get_y_fn):
    hist = {}
    for region in snapshot.regions:
        xval = get_x_fn(region, fmt)
        if not xval in hist:
            hist[xval] = 0
        hist[xval] += get_y_fn(region, fmt)
    return hist

def get_sorted_ranged_historgram(
        hist, fmt, fmt_x_fn, parse_x_fn, fmt_y_fn, y_aggr_fn):
    metrics = sorted(hist.keys())
    hist2 = []
    if fmt.hist_ranges is not None:
        hist_ranges = []
        for i in range(0, len(fmt.hist_ranges), 2):
            hist_ranges.append(
                    [parse_x_fn(fmt.hist_ranges[i]),
                     parse_x_fn(fmt.hist_ranges[i + 1])])
    else:
        min_metric, max_metric = metrics[0], metrics[-1]
        hist_ranges = get_hist_ranges(
                min_metric, max_metric, 10, fmt.hist_logscale,
                fmt.hist_cumulate)
    raw = fmt.raw_number
    for min_m, max_m in hist_ranges:
        yval = y_aggr_fn(
                [hist[x] for x in metrics if x >= min_m and x < max_m])
        metric_range_str = '[%s, %s)' % (
                fmt_x_fn(min_m, raw), fmt_x_fn(max_m, raw))
        yval_str = fmt_y_fn(yval, raw)
        hist2.append([metric_range_str, yval_str, yval])
    return hist2

def get_sz_region(region, fmt):
    return region.size()

def get_df_passed_sz_region(region):
    return region.sz_filter_passed

def sz_hist_str(snapshot, fmt, df_passed_sz, get_metric_fn, fmt_metric_fn,
                parse_metric_fn):
    if len(snapshot.regions) == 0:
        return 'no region in snapshot'
    if df_passed_sz is True:
        get_y_fn = get_df_passed_sz_region
    else:
        get_y_fn = get_sz_region
    hist = get_unsorted_histogram(snapshot, fmt, get_metric_fn, get_y_fn)
    hist2 = get_sorted_ranged_historgram(
            hist, fmt, fmt_metric_fn, parse_metric_fn, _damo_fmt_str.format_sz,
            lambda sz_list: sum(sz_list))

    return histogram_str(hist2)

def temperature_sz_hist_str(snapshot, record, fmt, df_passed_sz):
    def get_temperature(region, fmt):
        # set size weight zero
        weights = [0, fmt.temperature_weights[1], fmt.temperature_weights[2]]
        return temperature_of(region, weights)

    return sz_hist_str(snapshot, fmt, df_passed_sz, get_temperature,
                       _damo_fmt_str.format_nr, _damo_fmt_str.text_to_nr)

def get_last_used_time(region, fmt):
    if region.nr_accesses.percent > 0:
        return 0
    return region.age.usec

def recency_hist_str(snapshot, record, fmt, df_passed_sz):
    if len(snapshot.regions) == 0:
        return 'no region in snapshot'

    get_metric_fn = get_last_used_time
    fmt_metric_fn = _damo_fmt_str.format_time_us
    parse_metric_fn = _damo_fmt_str.text_to_us

    if df_passed_sz is True:
        get_y_fn = get_df_passed_sz_region
    else:
        get_y_fn = get_sz_region
    y_fmt_fn = _damo_fmt_str.format_sz
    y_aggr_fn = lambda vals: sum(vals)

    hist = get_unsorted_histogram(snapshot, fmt, get_metric_fn, get_y_fn)
    hist2 = get_sorted_ranged_historgram(
            hist, fmt, fmt_metric_fn, parse_metric_fn, y_fmt_fn, y_aggr_fn)

    return histogram_str(hist2)

class HistFns:
    get = None
    fmt = None
    parse = None

    def __init__(self, get, fmt, parse):
        self.get = get
        self.fmt = fmt
        self.parse = parse

hist_fns_map = {
        'recency': HistFns(get_last_used_time, _damo_fmt_str.format_time_us,
                           _damo_fmt_str.text_to_us),
        'sz': HistFns(get_sz_region, _damo_fmt_str.format_sz,
                      _damo_fmt_str.text_to_bytes),
        }

def temperature_str(region, raw, fmt):
    temperature = temperature_of(region, fmt.temperature_weights)
    return _damo_fmt_str.format_nr(temperature, raw)

def region_in(start, end, regions):
    for region in regions:
        if region.end <= start:
            continue
        if end <= region.start:
            break
        return True
    return False

def region_after(addr, regions):
    for region in regions:
        if region.start > addr:
            return region
    return None

class HeatPixel:
    start = None
    end = None
    total_heat = None
    temperature = None
    is_void = None

    def __init__(self, start, end, regions, temperature_weights):
        self.start = start
        self.end = end

        self.total_heat = 0
        self.is_void = True
        for region in regions:
            if region.end <= start:
                continue
            if end <= region.start:
                break
            self.is_void = False
            self.add_temperature(region, temperature_weights)
        self.temperature = self.total_heat / (self.end - self.start)

    def add_temperature(self, region, weights):
        start = self.start
        end = self.end
        # caller should ensure region intersects with self
        if start <= region.start and region.end <= end:
            # region is in the pixel
            pass
        elif region.start <= start and end <= region.end:
            # pixel is in the region
            region = copy.deepcopy(region)
            region.start = start
            region.end = end
        elif region.start < start:
            # region intersecting right part
            # <region>
            #    <pixel>
            region = copy.deepcopy(region)
            region.start = start
        else:
            # region intersecting left part
            #    <region>
            # <pixel>
            region = copy.deepcopy(region)
            region.end = end
        self.total_heat += temperature_of(region, weights) * (region.size())

def heatmap_str(snapshot, record, fmt):
    if len(snapshot.regions) == 0:
        return 'n/a (no region)'
    raw = fmt.raw_number
    static = fmt.snapshot_heatmap_static
    if static:
        total_sz = snapshot.regions[-1].end - snapshot.regions[0].start
    else:
        total_sz = 0
        for region in snapshot.regions:
            total_sz += region.size()
    map_length = fmt.snapshot_heatmap_width
    sz_unit = total_sz / map_length

    start = snapshot.regions[0].start
    pixels = []
    min_temperature = None
    max_temperature = None
    while start < snapshot.regions[-1].end:
        end = start + sz_unit
        pixels.append(HeatPixel(
            start, end, snapshot.regions, fmt.temperature_weights))
        if pixels[-1].is_void is True:
            next_region = region_after(end, snapshot.regions)
            if next_region is None:
                break
            start = next_region.start // sz_unit * sz_unit
            pixels[-1].end = start
        else:
            pixel = pixels[-1]
            if min_temperature is None or pixel.temperature < min_temperature:
                min_temperature = pixel.temperature
            if max_temperature is None or pixel.temperature > max_temperature:
                max_temperature = pixel.temperature
            start = end

    max_color_level = _damo_ascii_color.max_color_level()
    temperature_unit = (max_temperature - min_temperature) / max_color_level
    # single region?
    if temperature_unit == 0:
        temperature_unit = max_temperature / max_color_level
    if temperature_unit == 0:
        temperature_unit = 1
    dots = []
    for pixel in pixels:
        if pixel.is_void is True:
            nr_dots = (pixel.end - pixel.start) / sz_unit
            if not static and nr_dots > 4:
                dots.append('[...]')
            else:
                dots += ['.'] * int(nr_dots)
            continue
        temp_level = int(
                (pixel.temperature - min_temperature) / temperature_unit)
        dots.append(_damo_ascii_color.colored(
            '%d' % temp_level, fmt.snapshot_heatmap_colorset, temp_level))
    dots = ''.join(dots)
    comment = '# min/max temperatures: %s, %s, column size: %s' % (
            _damo_fmt_str.format_nr(min_temperature, raw),
            _damo_fmt_str.format_nr(max_temperature, raw),
            _damo_fmt_str.format_sz(sz_unit, raw))
    return '%s\n%s' % (dots, comment)

def df_passed_heatmap_str(snapshot, record, fmt):
    regions = []
    for region in snapshot.regions:
        cp_region = copy.deepcopy(region)
        if region.sz_filter_passed == 0:
            continue
        regions.append(cp_region)
    return heatmap_str(
            _damo_records.DamonSnapshot(
                snapshot.start_time, snapshot.end_time, regions,
                total_bytes=None),
            record, fmt)

def rescale(val, orig_scale_minmax, new_scale_minmax, logscale=True):
    '''Return a value in new scale

    Parameters
    ----------
    val : int, float
        The value to rescale
    orig_scale_minmax : list
        min/max values of original scale
    new_scale_minmax : list
        min/max values of new scale
    logscale : bool
        whether to use logscale (True) or linearscale (False)

    Returns
    -------
    float
        The rescaled value
    '''

    if logscale:
        log_val = math.log(val, 2) if val > 0 else 0
        log_minmax = [math.log(v, 2) if v > 0 else 0
                for v in orig_scale_minmax]
        val = log_val
        orig_scale_minmax = log_minmax
    orig_length = orig_scale_minmax[1] - orig_scale_minmax[0]
    new_length = new_scale_minmax[1] - new_scale_minmax[0]
    ratio = new_length / orig_length if orig_length > 0 else 1
    return (val - orig_scale_minmax[0]) * ratio + new_scale_minmax[0]

class BoxValue:
    val = None
    val_min_max = None      # list of min/max
    display_min_max = None  # list of min/max
    display_logscale = None # bool

    def __init__(self, val, val_min_max, display_min_max, display_logscale):
        self.val = val
        self.val_min_max = val_min_max
        self.display_min_max = display_min_max
        self.display_logscale = display_logscale

    def display_value(self):
        return rescale(self.val, self.val_min_max, self.display_min_max,
                self.display_logscale)

class ColoredBox:
    # BoxValue
    length = None
    color = None
    height = None

    colorset = None
    horizontal_align = None

    def __init__(self, length, horizontal_align, color, colorset, height):
        self.length = length
        self.color = color
        self.horizontal_align = horizontal_align
        self.colorset = colorset
        self.height = height

    def __str__(self):
        length = int(self.length.display_value())

        if self.height.val != None:
            height = int(self.height.display_value())
        else:
            self.height.display_min, self.height.display_max = [1, 1]
            height = 1

        if type(self.color.val) == str:
            row = '%s' % (self.color.val * length)
        else:
            color_level = int(self.color.display_value())
            row = '%s' % _damo_ascii_color.colored(
                    ('%d' % color_level) * length, self.colorset, color_level)
        row = '|%s|' % row

        if self.horizontal_align == 'right':
            indent = ' ' * (self.length.display_min_max[1] - length)
            row = indent + row

        box = '\n'.join([row] * height)
        if self.height.display_min_max[1] > 1:
            box += '\n'
        return box

class SortedAccessPatterns:
    sz_regions = None
    access_rates_percent = None
    ages_us = None

    def __init__(self, records):
        self.sz_regions = []
        self.access_rates_percent = []
        self.ages_us = []

        for record in records:
            for snapshot in record.snapshots:
                for region in snapshot.regions:
                    self.sz_regions.append(region.size())

                    region.nr_accesses.add_unset_unit(record.intervals)
                    self.access_rates_percent.append(
                            region.nr_accesses.percent)

                    region.age.add_unset_unit(record.intervals)
                    self.ages_us.append(region.age.usec)
        self.sz_regions.sort()
        self.access_rates_percent.sort()
        self.ages_us.sort()

class RegionBoxAttr:
    value_name = None
    display_min_max = None
    display_logscale = None

    def __init__(self, value_name, display_min_max, display_logscale):
        self.value_name = value_name
        self.display_min_max = display_min_max
        self.display_logscale = display_logscale

class RegionBoxFormat:
    sorted_access_patterns = None
    length = None
    horizontal_align = None
    color = None
    colorset = None
    height = None

    def __init__(self, sorted_access_patterns, length, horizontal_align, color,
                 colorset, height):
        self.sorted_access_patterns = sorted_access_patterns
        self.length = length
        self.horizontal_align = horizontal_align
        self.color = color
        self.colorset = colorset
        self.height = height

    def minmax(self, value_name):
        if value_name == 'size':
            sorted_vals = self.sorted_access_patterns.sz_regions
        elif value_name == 'access_rate':
            sorted_vals = self.sorted_access_patterns.access_rates_percent
        elif value_name == 'age':
            sorted_vals = self.sorted_access_patterns.ages_us
        return sorted_vals[0], sorted_vals[-1]

    def val_minmax(self, region, value_name):
        minval, maxval = self.minmax(value_name)
        if value_name == 'size':
            return region.size(), [minval, maxval]
        elif value_name == 'access_rate':
            return region.nr_accesses.percent, [minval, maxval]
        elif value_name == 'age':
            return region.age.usec, [minval, maxval]
        return None, None

    def to_str(self, region):
        length_val, length_val_minmax = self.val_minmax(region,
                self.length.value_name)
        color_val, color_val_minmax = self.val_minmax(region,
                self.color.value_name)
        if color_val == None:
            color_val = '-'
        height_val, height_val_minmax = self.val_minmax(region,
                self.height.value_name)

        box = '%s' % ColoredBox(
                BoxValue(length_val, length_val_minmax,
                    self.length.display_min_max, self.length.display_logscale),
                self.horizontal_align,
                BoxValue(color_val, color_val_minmax, [0, 9],
                    self.color.display_logscale),
                self.colorset,
                BoxValue(height_val, height_val_minmax,
                    self.height.display_min_max, self.height.display_logscale))
        return box

    def format_min_max(self, minval, maxval, value_name, raw):
        if value_name == 'size':
            return '[%s, %s]' % (_damo_fmt_str.format_sz(minval, raw),
                    _damo_fmt_str.format_sz(maxval, raw))
        if value_name == 'age':
            return '[%s, %s] ' % (_damo_fmt_str.format_time_us(minval, raw),
                    _damo_fmt_str.format_time_us(maxval, raw))
        if value_name == 'access_rate':
            return '[%s, %s]' % (_damo_fmt_str.format_percent(minval, raw),
                    _damo_fmt_str.format_percent(maxval, raw))

    def description_msg(self, raw):
        lines = []
        minval, maxval = self.minmax(self.length.value_name)
        lines.append('# length: %s (represents %s with [%d, %d] columns in %s)'
                % (self.length.value_name,
                    self.format_min_max(minval, maxval,
                        self.length.value_name, raw),
                    self.length.display_min_max[0],
                    self.length.display_min_max[1],
                    'logscale'
                    if self.length.display_logscale else 'linearscale'))

        minval, maxval = self.minmax(self.color.value_name)
        lines.append('# color: %s (represents %s with [%d, %d] number and color (%s) in %s)'
                % (self.color.value_name,
                    self.format_min_max(minval, maxval,
                        self.color.value_name, raw),
                    self.color.display_min_max[0],
                    self.color.display_min_max[1],
                    _damo_ascii_color.color_samples(self.colorset),
                    'logscale'
                    if self.color.display_logscale else 'linearscale'))

        minval, maxval = self.minmax(self.height.value_name)
        lines.append('# height: %s (represents %s with [%d, %d] rows in %s)'
                % (self.height.value_name,
                    self.format_min_max(minval, maxval,
                        self.height.value_name, raw),
                    self.height.display_min_max[0],
                    self.height.display_min_max[1],
                    'logscale'
                    if self.height.display_logscale else 'linearscale'))
        return '\n'.join(lines)

def apply_min_chars(min_chars, field_name, txt):
    # min_chars: [[<field name>, <number of min chars>]...]
    for name, nr in min_chars:
        try:
            nr = int(nr)
        except:
            print('wrong min_chars: %s' % min_chars)

        if name == field_name:
            if len(txt) >= nr:
                return txt
            txt += ' ' * (nr - len(txt))
            return txt
    return txt

def format_output(template, formatters, fmt, record, snapshot=None,
                  region=None, index=None):
    if template == '':
        return
    for formatter in formatters:
        if template.find(formatter.keyword) == -1:
            continue
        if formatters == record_formatters:
            txt = formatter.format_fn(record, fmt)
        elif formatters == snapshot_formatters:
            txt = formatter.format_fn(snapshot, record, fmt)
        elif formatters == region_formatters:
            txt = formatter.format_fn(index, region, fmt)
        txt = apply_min_chars(fmt.min_chars_for, formatter.keyword, txt)
        template = template.replace(formatter.keyword, txt)
    template = template.replace('\\n', '\n')
    return template


def temperature_of(region, weights):
    sz_weight, access_rate_weight, age_weight = weights
    sz_score = region.size() * sz_weight
    ar_score = region.nr_accesses.percent * access_rate_weight
    age_score = region.age.usec * age_weight
    score = sz_score + ar_score + age_score

    if region.nr_accesses.percent > 0:
        return score
    else:
        return -1 * score

def sorted_regions(regions, sort_fields, sort_dsc_keys, temperature_weights):
    for field in sort_fields:
        dsc = sort_dsc_keys != None and ('all' in sort_dsc_keys or
                                         field in sort_dsc_keys)
        if field == 'address':
            regions = sorted(regions, key=lambda r: r.start, reverse=dsc)
        elif field == 'access_rate':
            regions = sorted(regions, key=lambda r: r.nr_accesses.percent,
                    reverse=dsc)
        elif field == 'age':
            regions = sorted(regions, key=lambda r: r.age.usec, reverse=dsc)
        elif field == 'size':
            regions = sorted(regions, key=lambda r: r.size(), reverse=dsc)
        elif field == 'temperature':
            regions = sorted(
                    regions, reverse=dsc,
                    key=lambda r: temperature_of(r, temperature_weights))
    return regions

def fmt_records(fmt, records):
    sorted_access_patterns = SortedAccessPatterns(records)
    fmt.region_box_format = RegionBoxFormat(sorted_access_patterns,
            RegionBoxAttr(fmt.region_box_values[0],
                fmt.region_box_min_max_length,
                fmt.region_box_scales[0] == 'log'), fmt.region_box_align,
            RegionBoxAttr(fmt.region_box_values[1],
                [0, 9], fmt.region_box_scales[1] == 'log'),
            fmt.region_box_colorset,
            RegionBoxAttr(fmt.region_box_values[2],
                fmt.region_box_min_max_height,
                fmt.region_box_scales[2] == 'log'))

    outputs = []
    for record in records:
        outputs.append(
                format_output(
                    fmt.format_record_head, record_formatters, fmt, record))
        snapshots = record.snapshots

        for sidx, snapshot in enumerate(snapshots):
            outputs.append(
                    format_output(
                        fmt.format_snapshot_head, snapshot_formatters,
                        fmt, record, snapshot))
            for r in snapshot.regions:
                r.nr_accesses.add_unset_unit(record.intervals)
                r.age.add_unset_unit(record.intervals)
            for idx, r in enumerate(
                    sorted_regions(snapshot.regions, fmt.sort_regions_by,
                        fmt.sort_regions_dsc, fmt.temperature_weights)):
                outputs.append(
                        format_output(
                            fmt.format_region, region_formatters,
                            fmt, record, snapshot, r, idx))
            outputs.append(
                    format_output(
                        fmt.format_snapshot_tail, snapshot_formatters,
                        fmt, record, snapshot))

            if sidx < len(snapshots) - 1 and not fmt.total_sz_only():
                outputs.append('')
        outputs.append(
                format_output(
                    fmt.format_record_tail, record_formatters, fmt, record))
    outputs = [o for o in outputs if o is not None]
    return '\n'.join(outputs)

def pr_records_raw_form(records, raw_number):
    lines = []
    for record in records:
        snapshots = record.snapshots
        if len(snapshots) == 0:
            continue

        base_time = snapshots[0].start_time
        lines.append('base_time_absolute: %s\n' %
                _damo_fmt_str.format_time_ns(base_time, raw_number))

        for snapshot in snapshots:
            lines.append('monitoring_start:    %16s' %
                    _damo_fmt_str.format_time_ns(
                        snapshot.start_time - base_time, raw_number))
            lines.append('monitoring_end:      %16s' %
                    _damo_fmt_str.format_time_ns(
                        snapshot.end_time - base_time, raw_number))
            lines.append('monitoring_duration: %16s' %
                    _damo_fmt_str.format_time_ns(
                        snapshot.end_time - snapshot.start_time,
                        raw_number))
            lines.append('target_id: %s' % record.target_id)
            lines.append('nr_regions: %s' % len(snapshot.regions))
            lines.append('# %10s %12s  %12s  %11s %5s' %
                    ('start_addr', 'end_addr', 'length', 'nr_accesses', 'age'))
            for r in snapshot.regions:
                lines.append("%012x-%012x (%12s) %11d %5d" %
                        (r.start, r.end,
                            _damo_fmt_str.format_sz(r.size(), raw_number),
                            r.nr_accesses.samples, r.age.aggr_intervals
                                if r.age.aggr_intervals != None else -1))
            lines.append('')
    lines.append('')
    _damo_print.pr_with_pager_if_needed('\n'.join(lines))

def pr_records(fmt, records, dont_use_pager):
    if fmt.json:
        _damo_print.pr_with_pager_if_needed(
                json.dumps([r.to_kvpairs(fmt.raw_number) for r in records],
                           indent=4))
    elif fmt.raw:
        pr_records_raw_form(records, fmt.raw_number)
    else:
        to_show = fmt_records(fmt, records)
        if dont_use_pager:
            print(to_show)
        else:
            _damo_print.pr_with_pager_if_needed(fmt_records(fmt, records))

class ReportFormat:
    sort_regions_by = None
    sort_regions_dsc = None
    temperature_weights = None
    dont_merge_regions = None

    hist_logscale = None
    hist_cumulate = None
    hist_ranges = None

    format_record_head = None
    format_record_tail = None
    format_snapshot_head = None
    format_snapshot_tail = None
    format_region = None

    snapshot_heatmap_width = None
    snapshot_heatmap_colorset = None
    snapshot_heatmap_static = None

    region_box_values = None
    region_box_min_max_height = None
    region_box_min_max_length = None
    region_box_colorset = None
    region_box_scales = None
    region_box_align = None

    # RegionBoxFormat.  Set on fmt_records()
    region_box_format = None

    min_chars_for = None
    raw_number = None
    json = None
    raw = None


    @classmethod
    def from_args(cls, args):
        self = cls()
        self.sort_regions_by = args.sort_regions_by
        self.sort_regions_dsc = args.sort_regions_dsc
        self.temperature_weights = args.temperature_weights
        self.dont_merge_regions = args.dont_merge_regions
        self.hist_logscale = args.hist_logscale
        self.hist_cumulate = args.hist_cumulate
        self.hist_ranges = args.hist_ranges
        self.format_record_head = args.format_record_head
        self.format_record_tail = args.format_record_tail
        self.format_snapshot_head = args.format_snapshot_head
        self.format_snapshot_tail = args.format_snapshot_tail
        self.format_region = args.format_region
        self.snapshot_heatmap_width = args.snapshot_heatmap_width
        self.snapshot_heatmap_static = args.snapshot_heatmap_static
        self.snapshot_heatmap_colorset = args.snapshot_heatmap_colorset
        self.region_box_values = args.region_box_values
        self.region_box_min_max_height = args.region_box_min_max_height
        self.region_box_min_max_length = args.region_box_min_max_length
        self.region_box_colorset = args.region_box_colorset
        self.region_box_scales = args.region_box_scales
        self.region_box_align = args.region_box_align
        self.min_chars_for = args.min_chars_for
        self.raw_number = args.raw_number
        self.json = args.json
        self.raw = args.raw_form
        return self

    def total_sz_only(self):
        return (
                self.format_snapshot_head == '' and
                self.format_region == '' and
                self.format_snapshot_tail == '<total bytes>')

    def to_kvpairs(self, raw):
        return {
                'sort_regions_by': self.sort_regions_by,
                'sort_regions_dsc': self.sort_regions_dsc,
                'temperature_weights': self.temperature_weights,
                'dont_merge_regions': self.dont_merge_regions,
                'hist_logscale': self.hist_logscale,
                'hist_cumulate': self.hist_cumulate,
                'hist_ranges': self.hist_ranges,
                'format_record_head': self.format_record_head,
                'format_record_tail': self.format_record_tail,
                'format_snapshot_head': self.format_snapshot_head,
                'format_snapshot_tail': self.format_snapshot_tail,
                'format_region': self.format_region,
                'snapshot_heatmap_width': self.snapshot_heatmap_width,
                'snapshot_heatmap_static': self.snapshot_heatmap_static,
                'snapshot_heatmap_colorset': self.snapshot_heatmap_colorset,
                'region_box_values': self.region_box_values,
                'region_box_min_max_height': self.region_box_min_max_height,
                'region_box_min_max_length': self.region_box_min_max_length,
                'region_box_colorset': self.region_box_colorset,
                'region_box_scales': self.region_box_scales,
                'region_box_align': self.region_box_align,
                'min_chars_for': self.min_chars_for,
                'raw_number': self.raw_number,
                'json': self.json,
                }

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls()
        self.sort_regions_by = kvpairs['sort_regions_by']
        self.sort_regions_dsc = kvpairs['sort_regions_dsc']
        self.temperature_weights = kvpairs['temperature_weights']
        self.dont_merge_regions = kvpairs['dont_merge_regions']
        # hist_logscale introduced after v2.7.0
        if 'hist_logscale' in kvpairs:
            self.hist_logscale = kvpairs['hist_logscale']
        else:
            self.hist_logscale = False
        # hist_cumulate introduced after v2.7.6
        if 'hist_cumulate' in kvpairs:
            self.hist_cumulate = kvpairs['hist_cumulate']
        else:
            self.hist_cumulate = False
        # hist_ranges introduced after v2.7.5
        if 'hist_ranges' in kvpairs:
            self.hist_ranges = kvpairs['hist_ranges']
        else:
            self.hist_ranges = None
        self.format_record_head = kvpairs['format_record_head']
        self.format_record_tail = kvpairs['format_record_tail']
        self.format_snapshot_head = kvpairs['format_snapshot_head']
        self.format_snapshot_tail = kvpairs['format_snapshot_tail']
        self.format_region = kvpairs['format_region']
        # snapshot_heatmap_width introduced from v2.5.6
        if 'snapshot_heatmap_width' in kvpairs:
            self.snapshot_heatmap_width = kvpairs['snapshot_heatmap_width']
        else:
            self.snapshot_heatmap_width = 80
        # snapshot_heatmap_colorset introduced from v2.5.6
        if 'snapshot_heatmap_colorset' in kvpairs:
            self.snapshot_heatmap_width = kvpairs['snapshot_heatmap_width']
        else:
            self.snapshot_heatmap_width = 'gray'
        # snapshot_heatmap_static introduced after v2.6.7
        if 'snapshot_heatmap_static' in kvpairs:
            self.snapshot_heatmap_static = kvpairs['snapshot_heatmap_static']
        else:
            self.snapshot_heatmap_static = False
        self.region_box_values = kvpairs['region_box_values']
        self.region_box_min_max_height = kvpairs['region_box_min_max_height']
        self.region_box_min_max_length = kvpairs['region_box_min_max_length']
        self.region_box_colorset = kvpairs['region_box_colorset']
        self.region_box_scales = kvpairs['region_box_scales']
        self.region_box_align = kvpairs['region_box_align']
        self.min_chars_for = kvpairs['min_chars_for']
        self.raw_number = kvpairs['raw_number']
        self.json = kvpairs['json']
        return self

def has_ops_filters(records):
    for record in records:
        for scheme_filter in record.scheme_filters:
            if scheme_filter.handled_by_ops():
                return True
    return False

def set_formats_hist_style(args, fmt, records):
    if args.style == 'temperature-sz-hist':
        legend = '<temperature>'
        hist_keyword = '<temperature-sz histogram>'
        filter_passed_hist_keyword = '<temperature-df-passed-sz histogram>'
    else:
        # args.style == 'recency-sz-hist':
        legend = '<last accessed time (us)>'
        hist_keyword = '<recency-sz histogram>'
        filter_passed_hist_keyword = '<recency-df-passed-sz histogram>'

    snapshot_head_content = []
    if has_ops_filters(records):
        snapshot_head_content += [
                '# damos filters (df): <filters passed type>',
                '%s <df-passed size>' % legend, filter_passed_hist_keyword, '']
    snapshot_head_content += ['%s <total size>' % legend, hist_keyword]
    fmt.format_snapshot_head = '\n'.join(snapshot_head_content)
    fmt.format_region = ''

def set_formats(args, records):
    fmt = ReportFormat.from_args(args)

    if args.format is not None:
        if len(args.format) == 1 and len(args.format[0]) == 1:
            fmt_string = args.format[0][0]
            if os.path.isfile(fmt_string):
                with open(fmt_string, 'r') as f:
                    fmt_string = f.read()
            try:
                return ReportFormat.from_kvpairs(json.loads(fmt_string)), None
            except:
                return None, 'wrong --format option'
        for format_fields in args.format:
            if len(format_fields) != 3:
                return None, 'wrong --format option (%s)' % format_fields
            action, target_area, fmt_string = format_fields
            if not action in ['set', 'append']:
                return None, 'wrong --format action (%s)' % format_fields
            if not target_area in ['record_head', 'snapshot_head', 'region',
                                   'snapshot_tail', 'record_tail']:
                return None, 'wrong --format target area (%s)' % format_fields
            if action == 'set':
                if target_area == 'record_head':
                    fmt.format_record_head = fmt_string
                elif target_area == 'snapshot_head':
                    fmt.format_snapshot_head = fmt_string
                elif target_area == 'region':
                    fmt.format_region = fmt_string
                elif target_area == 'snapshot_tail':
                    fmt.format_snapshot_tail = fmt_string
                elif target_area == 'record_tail':
                    fmt.format_record_tail = fmt_string

    if args.style == 'simple-boxes':
        fmt.format_snapshot_head = default_snapshot_head_format_without_heatmap
        fmt.format_region = '<box> size <size> access rate <access rate> age <age>'
        fmt.region_box_min_max_height = [1, 1]
        fmt.region_box_min_max_length = [1, 40]
        fmt.region_box_align = 'right'
        fmt.region_box_colorset = 'emotion'
    elif args.style in ['temperature-sz-hist', 'recency-sz-hist']:
        set_formats_hist_style(args, fmt, records)
    elif args.style == 'cold':
        fmt.format_region = '<box> <size> access <access rate> <age>'
        fmt.region_box_min_max_height = [1, 1]
        fmt.region_box_min_max_length = [1, 40]
        fmt.region_box_align = 'right'
        fmt.region_box_colorset = 'emotion'
        fmt.sort_regions_by = ['temperature']
    elif args.style == 'hot':
        fmt.format_region = '<box> <size> access <access rate> <age>'
        fmt.region_box_min_max_height = [1, 1]
        fmt.region_box_min_max_length = [1, 40]
        fmt.region_box_align = 'right'
        fmt.region_box_colorset = 'emotion'
        fmt.sort_regions_by = ['temperature']
        fmt.sort_regions_dsc = ['temperature']

    fmt.region_box_values = [v if v != 'none' else None
            for v in args.region_box_values]

    if args.total_sz_only:
        fmt.format_snapshot_head = ''
        fmt.format_region = ''
        fmt.format_snapshot_tail = '<total bytes>'

    if args.region_box:
        if fmt.region_box_min_max_height[1] > 1:
            fmt.format_region = '<box>%s' % default_region_format
        else:
            fmt.format_region = '<box>\n%s' % default_region_format
        if fmt.format_snapshot_tail.find('<region box description>') == -1:
            fmt.format_snapshot_tail = ('%s\n<region box description>' %
                    fmt.format_record_tail)
    if len(records) == 0:
        return fmt, None

    if fmt.format_record_head == None:
        if len(records) > 1:
            fmt.format_record_head = default_record_head_format
        else:
            fmt.format_record_head = ''

    if fmt.format_region == default_region_format:
        for record in records:
            if len(record.scheme_filters) > 0:
                fmt.format_region += ' df-passed <filters passed bytes>'
                break

    ops_filters_installed = False
    intervals_tuning_enabled = False
    for record in records:
        if len(record.scheme_filters) > 0:
            ops_filters_installed = True
        if record.intervals is not None:
            if record.intervals.intervals_goal.enabled():
                intervals_tuning_enabled = True
        if ops_filters_installed and intervals_tuning_enabled:
            break

    if fmt.format_snapshot_tail == default_snapshot_tail_format:
        if ops_filters_installed:
            fmt.format_snapshot_tail = default_snapshot_tail_format_filter_installed
        # further check if scheme action is not stat
        if args.tried_regions_of is not None:
            fmt.format_snapshot_tail += '\nscheme stats\n<damos stats>'

    if fmt.format_snapshot_head == None:
        need_snapshot_head = False
        for record in records:
            if len(record.snapshots) > 1:
                need_snapshot_head = True
                break
        if need_snapshot_head:
            fmt.format_snapshot_head = default_snapshot_head_format
        else:
            fmt.format_snapshot_head = 'heatmap: <heatmap>'
    if '<filters passed bytes>' in fmt.format_region:
        fmt.format_snapshot_head += '\n# damos filters (df): <filters passed type>'
        fmt.format_snapshot_head += '\ndf-pass: <filters passed heatmap>'

    if fmt.format_record_tail == '' and intervals_tuning_enabled:
        fmt.format_record_tail = 'monitoring intervals: <intervals>'

    if args.format is not None:
        for action, target_area, fmt_String in args.format:
            if action == 'append':
                if target_area == 'record_head':
                    fmt.format_record_head += fmt_string
                elif target_area == 'snapshot_head':
                    fmt.format_snapshot_head += fmt_string
                elif target_area == 'region':
                    fmt.format_region += fmt_string
                elif target_area == 'snapshot_tail':
                    fmt.format_snapshot_tail += fmt_string
                elif target_area == 'record_tail':
                    fmt.format_record_tail += fmt_string

    return fmt, None

def handle_ls_keywords(args):
    if args.ls_record_format_keywords:
        print("use 'damo help access_format_options record_format_keywords' instead")
        return True
    if args.ls_snapshot_format_keywords:
        print("use 'damo help access_format_options snapshot_format_keywords' instead")
        return True
    if args.ls_region_format_keywords:
        print("use 'damo help access_format_options region_format_keywords' instead")
        return True
    return False

class CacheSpec:
    size = None
    ways = None
    sz_line = None
    nr_cache_sets = None

    def __init__(self, size, ways, sz_line):
        self.size = size
        self.ways = ways
        self.sz_line = sz_line

        self.nr_cache_sets = self.size / self.ways / self.sz_line

    def set_for_pa(self, addr):
        return addr / self.sz_line % self.nr_cache_sets

def translate_regions_to_cache_space(regions, cache_spec):
    set_regions = []
    for i in range(int(cache_spec.nr_cache_sets)):
        set_regions.append(_damon.DamonRegion(
            start=i, end=i+1,
            nr_accesses=0, nr_accesses_unit=_damon.unit_samples,
            age=0, age_unit=_damon.unit_aggr_intervals))
    for region in regions:
        for addr in range(region.start, region.end, cache_spec.sz_line):
            csidx = int(cache_spec.set_for_pa(addr))
            if csidx >= len(set_regions):
                print(csidx, len(set_regions))
                exit(0)
            set_regions[csidx].nr_accesses.samples += region.nr_accesses.samples
    converted = []
    for set_region in set_regions:
        if len(converted) == 0:
            converted.append(set_region)
            continue
        if converted[-1].nr_accesses.samples == set_region.nr_accesses.samples:
            converted[-1].end = set_region.end
        else:
            converted.append(set_region)
    return converted

def translate_records_to_cache_space(records, cs, cw, cl):
    cache_spec = CacheSpec(cs, cw, cl)
    for record in records:
        for snapshot in record.snapshots:
            snapshot.regions = translate_regions_to_cache_space(
                    snapshot.regions, cache_spec)

signal_received = False

def sighandler(signum, frame):
    global signal_received
    print('\nsignal %s received' % signum)
    signal_received = True

def recursive_del(dirpath):
    for filename in os.listdir(dirpath):
        filepath = os.path.join(dirpath, filename)
        if os.path.isdir(filepath):
            recursive_del(filepath)
        else:
            os.remove(filepath)
    os.rmdir(dirpath)

def handle_exec(cmd, records):
    if cmd == 'interpreter':
        code.interact(
                local=locals(),
                banner="DAMON records are available as 'records'",
                exitmsg='Exit the session.')
        return None
    cmd_fields = cmd.split()
    if not os.path.isfile(cmd_fields[0]):
        return 'first token is not file'
    script_dir_path = tempfile.mkdtemp(prefix='damo_script_')
    script_module_name = 'damo_report_access_script'
    script_file_name = '%s.py' % script_module_name
    with open(cmd_fields[0], 'r') as f:
        with open(os.path.join(script_dir_path, script_file_name), 'w') as f2:
            f2.write(f.read())
    sys.path.append(script_dir_path)
    import damo_report_access_script
    damo_report_access_script.main(records, cmd_fields)
    recursive_del(script_dir_path)
    return None

def read_and_show(args):
    record_filter, err = _damo_records.args_to_filter(args)
    if err != None:
        print(err)
        exit(1)

    if args.input_file == None:
        _damon.ensure_root_and_initialized(args)
        if _damon.any_kdamond_running() is False:
            if os.path.exists('damon.data'):
                args.input_file = 'damon.data'

    dfilters, err = _damon_args.damos_options_to_filters(
            args.snapshot_damos_filter)
    if err is not None:
        print('wrong --snapshot_damos_filter (%s)' % err)
        exit(1)

    if args.repeat is None:
        repeat_delay = 0
        repeat_count = 1
    elif len(args.repeat) == 0:
        repeat_delay = 1
        repeat_count = -1
    elif len(args.repeat) == 2:
        repeat_delay = _damo_fmt_str.text_to_sec(args.repeat[0])
        repeat_count = _damo_fmt_str.text_to_nr(args.repeat[1])
    else:
        print('--repeat receives only zero or two arguments')
        exit(1)

    signal.signal(signal.SIGINT, sighandler)
    global signal_received

    read_show_count = 0
    while read_show_count < repeat_count or repeat_count == -1:
        if signal_received is True:
            break
        records, err = _damo_records.get_records(
                    tried_regions_of=args.tried_regions_of,
                    record_file=args.input_file, snapshot_damos_filters=dfilters,
                    record_filter=record_filter,
                    total_sz_only=args.total_sz_only,
                    dont_merge_regions=args.dont_merge_regions)
        if err != None:
            print(err)
            exit(1)
        if signal_received is True:
            break

        if args.exec:
            err = handle_exec(args.exec, records)
            if err is not None:
                print('format_script handling fail (%s)' % err)
                exit(1)
            time.sleep(repeat_delay)
            read_show_count += 1
            continue

        if len([r for r in records if r.intervals is None]) != 0:
            if not args.json and not args.raw_form:
                print('some records lack the intervals information')
                exit(1)

        fmt, err = set_formats(args, records)
        if err is not None:
            print('format setting failed (%s)' % err)

        if args.on_cache is not None:
            sz_cache = _damo_fmt_str.text_to_bytes(args.on_cache[0])
            ways_cache = _damo_fmt_str.text_to_nr(args.on_cache[1])
            sz_cache_line = _damo_fmt_str.text_to_bytes(args.on_cache[2])
            translate_records_to_cache_space(
                    records, sz_cache, ways_cache, sz_cache_line)

        for record in records:
            try:
                pr_records(fmt, records,
                           dont_use_pager = args.repeat is not None)
            except BrokenPipeError as e:
                # maybe user piped to 'less' like pager, and quit from it
                pass

        time.sleep(repeat_delay)
        read_show_count += 1

def main(args):
    handled = handle_ls_keywords(args)
    if handled:
        return

    read_and_show(args)

def add_fmt_args(parser, hide_help=False):
    # how to show, in simple selection
    parser.add_argument(
            '--style', choices=['detailed', 'simple-boxes',
                                'temperature-sz-hist', 'recency-sz-hist',
                                'cold', 'hot'],
            default='detailed',
            help='output format selection among pre-configures ones')
    # how to show, in highly tunable way
    parser.add_argument(
            '--sort_regions_by', nargs='+',
            choices=['address', 'access_rate', 'age', 'size', 'temperature'],
            default=['address'],
            help='fields to sort regions by'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--sort_regions_dsc',
            choices=['address', 'access_rate', 'age', 'size', 'temperature',
                     'all'],
            nargs='+',
            help='sort regions in descending order for the given keys'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--temperature_weights', type=int, metavar='<int>', nargs=3,
            default=[0, 100, 100],
            help=' '.join(
                ['temperature weights for size, access rate, and age',
                 'of each region']) if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--hist_logscale', action='store_true',
                        help='draw histograms in logscale'
                        if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--hist_ranges', nargs='+',
                        help='histogram range values'
                        if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--hist_cumulate', action='store_true',
                        help='draw histogram in cumulative way')

    fmt_flexible_metavar='<set|append> <record_head|snapshot_head|region|snapshot_tail|record_tail> <format string>'

    parser.add_argument('--format', metavar='<%s> or <json string>' %
                        fmt_flexible_metavar, action='append', nargs='+',
                        help='visualization format in json format')

    # don't set default for record head and snapshot head because it depends on
    # given number of record and snapshots.  Decide those in set_formats().
    # below five options can be replaced with --format.  Keep backward
    # compatibility support, but hide from help messages.
    parser.add_argument(
            '--format_record_head', metavar='<template>',
            # help='output format to show at the beginning of each record'
            help=argparse.SUPPRESS)
    parser.add_argument('--format_record_tail', metavar='<template>',
                        default='',
                        # help='output format to show at the end of each record'
                        help=argparse.SUPPRESS)
    parser.add_argument(
            '--format_snapshot_head', metavar='<template>',
            # help='output format to show at the beginning of each snapshot'
            help=argparse.SUPPRESS)
    parser.add_argument(
            '--format_snapshot_tail', metavar='<template>',
            default=default_snapshot_tail_format,
            # help='output format to show at the end of each snapshot'
            help=argparse.SUPPRESS)
    parser.add_argument('--format_region', metavar='<template>',
                        default=default_region_format,
                        # help='output format to show for each memory region'
                        help=argparse.SUPPRESS)

    # for snapshot heatmap
    parser.add_argument(
            '--snapshot_heatmap_width', metavar='<number>', default=80,
            type=int,
            help='width of snapshot heatmap'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--snapshot_heatmap_colorset', default='gray',
            choices=_damo_ascii_color.colorsets.keys(),
            help='snapshot heatmap colorset'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--snapshot_heatmap_static', action='store_true',
            help='draw snapshot heatmap as static'
            if not hide_help else argparse.SUPPRESS)

    # for region box
    parser.add_argument('--region_box', action='store_true',
            help='show region access pattern as a box'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_values',
            choices=['size', 'access_rate', 'age', 'none'], nargs=3,
            default=['age', 'access_rate', 'size'],
            help='values to show via the <box>\'s length, color, and height'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_min_max_length', nargs=2, type=int,
            metavar=('<min>', '<max>'), default=[1, 30],
            help='minimum and maximum number of the region box\' length'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_min_max_height', nargs=2, type=int,
            metavar=('<min>', '<max>'), default=[1, 5],
            help='minimum and maximum number of region box\' height'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_colorset', default='gray',
            choices=['gray', 'flame', 'emotion'],
            help='colorset to use for region box'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_scales', choices=['linear', 'log'],
            nargs=3, default=['log', 'linear', 'log'],
            help='scale of region box\' length, color, and height'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--region_box_align', choices=['left', 'right'], default='left',
            help='where to align the region box'
            if not hide_help else argparse.SUPPRESS)

    # misc control
    parser.add_argument(
            '--min_chars_for', nargs=2,
            metavar=('<keyword>', '<number>'), action='append',
            default=[['<index>', 3],
                     ['<start address>', 12],['<end address>', 11],
                     ['<size>', 11], ['<access rate>', 5],
                     ['<age>', 13],
                     ['<filters passed type>', 10],
                     ],
            help='minimum character for each keyword of the format'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--total_sz_only', action='store_true',
            help='print only total size of the regions for each snapshot')
    parser.add_argument('--dont_merge_regions', action='store_true',
            help='don\'t merge contiguous regions of same access pattern'
            if not hide_help else argparse.SUPPRESS)

    parser.add_argument('--raw_number', action='store_true',
            help='use machine-friendly raw numbers')
    parser.add_argument('--json', action='store_true',
            help='print in json format')
    parser.add_argument('--raw_form', action='store_true',
                        help='print in raw format'
                        if not hide_help else argparse.SUPPRESS)
    parser.add_argument('--ls_record_format_keywords', action='store_true',
                        # help='list available record format keywords'
                        help=argparse.SUPPRESS)
    parser.add_argument('--ls_snapshot_format_keywords', action='store_true',
                        # help='list available snapshot format keywords'
                        help=argparse.SUPPRESS)
    parser.add_argument('--ls_region_format_keywords', action='store_true',
                        # help='list available region format keywords'
                        help=argparse.SUPPRESS)
    if hide_help:
        if parser.epilog is None:
            parser.epilog = ''
        else:
            parser.epilog += ' '
        parser.epilog += ' '.join([
            "More options for customized format are available.",
            "Do 'damo help access_format_options -h' for those options.",
            ])

def set_argparser(parser):
    parser.description = 'Show DAMON-monitored access pattern'
    parser.epilog=' '.join([
        'If --input_file is not provided, capture snapshot.',
        'If --input_file is not provided, DAMON is not running,',
        'and "damon.data" file exists, use "damon.data" as --input_file.'])

    _damon_args.set_common_argparser(parser)

    # what to show
    _damo_records.set_filter_argparser(parser, hide_help=True)

    parser.add_argument('--input_file', metavar='<file>',
            help='source of the access pattern to show')
    parser.add_argument('--tried_regions_of', nargs=3, type=int,
            action='append',
            metavar=('<kdamond idx>', '<context idx>', '<scheme idx>'),
            help='show tried regions of given schemes')
    _damo_records.set_snapshot_damos_filters_option(parser)
    add_fmt_args(parser, hide_help=True)

    parser.add_argument('--exec', metavar='<command or \'interpreter\'>',
                        help='execute python code with the records')
    parser.add_argument(
            '--on_cache', nargs=3,
            metavar=('<cache size>', '<cache ways>', '<cache line size>'),
            # visualize access patterns on a virtual cache (EXPERIMENTAL)
            help=argparse.SUPPRESS)
    parser.add_argument(
            '--repeat', nargs='*', metavar=('<delay>', '<count>'),
            help='repeat <count> times with <delay> time interval')



# SPDX-License-Identifier: GPL-2.0

import argparse
import collections
import copy
import json
import os
import random
import signal
import subprocess
import time
import zlib

import _damo_fmt_str
import _damo_subproc
import _damon
import _damon_args
import damo_report_access

PERF = 'perf'
perf_event_damon_aggregated = 'damon:damon_aggregated'
perf_event_damos_before_apply = 'damon:damos_before_apply'

class DamonSnapshot:
    '''
    Contains a snapshot of data access monitoring results
    '''
    start_time = None
    end_time = None
    regions = None
    total_bytes = None
    damos_stats = None

    def update_total_bytes(self):
        self.total_bytes = sum([r.size() for r in self.regions])

    def __init__(self, start_time, end_time, regions, total_bytes, damos_stats=None):
        self.start_time = start_time
        self.end_time = end_time
        self.regions = regions
        self.total_bytes = total_bytes
        if self.total_bytes is None:
            self.update_total_bytes()
        self.damos_stats = damos_stats

    @classmethod
    def from_kvpairs(cls, kv):
        damos_stats = None
        if 'damos_stats' in kv and kv['damos_stats'] is not None:
            damos_stats = _damon.DamosStats.from_kvpairs(kv['damos_stats'])
        return DamonSnapshot(
                _damo_fmt_str.text_to_ns(kv['start_time']),
                _damo_fmt_str.text_to_ns(kv['end_time']),
                [_damon.DamonRegion.from_kvpairs(r) for r in kv['regions']],
                _damo_fmt_str.text_to_bytes(kv['total_bytes'])
                if 'total_bytes' in kv and kv['total_bytes'] is not None
                else None, damos_stats)

    def to_kvpairs(self, raw=False):
        damos_stats_kv = None
        if self.damos_stats is not None:
            damos_stats_kv = self.damos_stats.to_kvpairs(raw)
        return collections.OrderedDict([
            ('start_time', _damo_fmt_str.format_time_ns_exact(
                self.start_time, raw)),
            ('end_time', _damo_fmt_str.format_time_ns_exact(
                self.end_time, raw)),
            ('regions', [r.to_kvpairs() for r in self.regions]),
            ('total_bytes', _damo_fmt_str.format_sz(self.total_bytes, raw)
                if self.total_bytes is not None else None),
            ('damos_stats', damos_stats_kv)
            ])

class DamonRecord:
    '''
    Contains data access monitoring results for single target
    '''
    kdamond_idx = None
    context_idx = None
    intervals = None
    scheme_idx = None
    target_id = None
    scheme_filters = None
    snapshots = None

    def __init__(self, kd_idx, ctx_idx, intervals, scheme_idx, target_id,
                 scheme_filters):
        self.kdamond_idx = kd_idx
        self.context_idx = ctx_idx
        self.intervals = intervals
        self.scheme_idx = scheme_idx
        self.target_id = target_id
        self.scheme_filters = scheme_filters
        self.snapshots = []

    @classmethod
    def from_kvpairs(cls, kv):
        for keyword in ['kdamond_idx', 'context_idx', 'intervals',
                'scheme_idx']:
            if not keyword in kv:
                kv[keyword] = None
        if not 'scheme_filters' in kv:
            kv['scheme_filters'] = []

        record = DamonRecord(
                kv['kdamond_idx'], kv['context_idx'],
                _damon.DamonIntervals.from_kvpairs(kv['intervals'])
                if kv['intervals'] is not None else None,
                kv['scheme_idx'], kv['target_id'],
                [_damon.DamosFilter.from_kvpairs(pairs) for pairs in
                 kv['scheme_filters']])
        record.snapshots = [DamonSnapshot.from_kvpairs(s)
                for s in kv['snapshots']]

        return record

    def to_kvpairs(self, raw=False):
        ordered_dict = collections.OrderedDict()
        ordered_dict['kdamond_idx'] = self.kdamond_idx
        ordered_dict['context_idx'] = self.context_idx
        ordered_dict['intervals'] = (self.intervals.to_kvpairs(raw)
                if self.intervals is not None else None)
        ordered_dict['scheme_idx'] = self.scheme_idx
        ordered_dict['target_id'] = self.target_id
        if self.scheme_filters is not None:
            ordered_dict['scheme_filters'] = [
                    s.to_kvpairs(raw) for s in self.scheme_filters]
        else:
            ordered_dict['scheme_filters'] = []
        ordered_dict['snapshots'] = [s.to_kvpairs(raw) for s in self.snapshots]
        return ordered_dict

    def can_merge(self, other):
        return (
                self.kdamond_idx == other.kdamond_idx and
                self.context_idx == other.context_idx and
                self.intervals == other.intervals and
                self.scheme_idx == other.scheme_idx and
                self.target_id == other.target_id)

    def merge(self, other):
        self.snapshots += other.snapshots
        self.snapshots.sort(key=lambda s: s.start_time)

# for monitoring results manipulation

def merge_records(records):
    merged_records = []
    for record in records:
        merged = False
        for merged_record in merged_records:
            if merged_record.can_merge(record):
                merged_record.merge(record)
                merged = True
                break
        if not merged:
            merged_records.append(record)
    return merged_records

def regions_intersect(r1, r2):
    return not (r1.end <= r2.start or r2.end <= r1.start)

def sz_filter_passed_for_new_size(region, new_size):
    if region.sz_filter_passed is None:
        return None
    filter_pass_ratio = region.sz_filter_passed / region.size()
    return int(filter_pass_ratio * new_size)

def add_region(regions, region, nr_acc_to_add):
    for r in regions:
        if not regions_intersect(r, region):
            continue
        if not r in nr_acc_to_add:
            nr_acc_to_add[r] = 0
        nr_acc_to_add[r] = max(nr_acc_to_add[r],
                region.nr_accesses.samples)

        new_regions = []
        if region.start < r.start:
            new_regions.append(_damon.DamonRegion(
                region.start, r.start,
                region.nr_accesses.samples, _damon.unit_samples,
                region.age.aggr_intervals, _damon.unit_aggr_intervals,
                sz_filter_passed_for_new_size(
                    region, r.start - region.start)))
        if r.end < region.end:
            new_regions.append(_damon.DamonRegion(
                    r.end, region.end,
                    region.nr_accesses.samples, _damon.unit_samples,
                    region.age.aggr_intervals,
                    _damon.unit_aggr_intervals,
                    sz_filter_passed_for_new_size(
                        region, region.end - r.end)))

        for new_r in new_regions:
            add_region(regions, new_r, nr_acc_to_add)
        return
    regions.append(region)

def aggregate_snapshots(snapshots):
    new_regions = []
    for snapshot in snapshots:
        # Suppose the first snapshot has a region 1-10:5, and the second
        # snapshot has two regions, 1-5:2, 5-10: 4.  Aggregated snapshot should
        # be 1-10:9.  That is, we should add maximum nr_accesses of
        # intersecting regions.  nr_acc_to_add contains the information.
        nr_acc_to_add = {}
        for region in snapshot.regions:
            add_region(new_regions, region, nr_acc_to_add)
        for region in nr_acc_to_add:
            region.nr_accesses.samples += nr_acc_to_add[region]
            region.nr_accesses.val = region.nr_accesses.samples
            region.nr_accesses.unit = _damon.unit_samples

    new_snapshot = DamonSnapshot(snapshots[0].start_time,
            snapshots[-1].end_time, new_regions, None)
    return new_snapshot

def adjusted_snapshots(snapshots, aggregate_interval_us):
    adjusted = []
    to_aggregate = []
    for snapshot in snapshots:
        to_aggregate.append(snapshot)
        interval_ns = to_aggregate[-1].end_time - to_aggregate[0].start_time
        if interval_ns >= aggregate_interval_us * 1000:
            adjusted.append(aggregate_snapshots(to_aggregate))
            to_aggregate = []
    return adjusted

def adjust_records(records, aggregate_interval, nr_snapshots_to_skip):
    for record in records:
        if record.intervals is not None:
            record.intervals.aggr = aggregate_interval
        record.snapshots = adjusted_snapshots(
                record.snapshots[nr_snapshots_to_skip:], aggregate_interval)

# For reading monitoring results from a file

# if number of snapshots is one and the file type is record or perf script,
# write_damon_records() adds a fake snapshot for snapshot start time deduction.
def is_fake_snapshot(snapshot):
    if len(snapshot.regions) != 1:
        return False
    r = snapshot.regions[0]
    return (r.start == 0 and r.end == 0 and
            r.nr_accesses.samples == -1 and r.age.aggr_intervals == -1)

def set_first_snapshot_start_time(records):
    for record in records:
        snapshots = record.snapshots
        if len(snapshots) < 2:
            break
        end_time = snapshots[-1].end_time
        start_time = snapshots[0].end_time
        nr_snapshots = len(snapshots) - 1
        snapshot_time = float(end_time - start_time) / nr_snapshots
        snapshots[0].start_time = snapshots[0].end_time - snapshot_time

        if is_fake_snapshot(snapshots[-1]):
            del record.snapshots[-1]

def record_of(target_id, records, intervals):
    for record in records:
        if record.target_id == target_id:
            return record
    record = DamonRecord(None, None, intervals, None, target_id, scheme_filters=[])
    records.append(record)
    return record

def parse_damon_aggregated_perf_script_fields(fields):
    '''
    The line is like below:

    kdamond.0  4452 [000] 82877.315633: damon:damon_aggregated: \
            target_id=18446623435582458880 nr_regions=17 \
            140731667070976-140731668037632: 0 3

    Note that the last field is not in the early version[1].

    [1] https://lore.kernel.org/linux-mm/df8d52f1fb2f353a62ff34dc09fe99e32ca1f63f.1636610337.git.xhao@linux.alibaba.com/
    '''

    if not len(fields) in [9, 10]:
        return None, None, None, None

    end_time = int(float(fields[3][:-1]) * 1000000000)
    target_id = int(fields[5].split('=')[1])
    nr_regions = int(fields[6].split('=')[1])

    start_addr, end_addr = [int(x) for x in fields[7][:-1].split('-')]
    nr_accesses = int(fields[8])
    if len(fields) == 10:
        age = int(fields[9])
    else:
        age = None
    region = _damon.DamonRegion(start_addr, end_addr, nr_accesses,
            _damon.unit_samples, age, _damon.unit_aggr_intervals)

    return region, end_time, target_id, nr_regions

def parse_damos_before_apply_perf_script_fields(fields):
    '''
    The line is like below:

    kdamond.0  4452 [000] 82877.315633: damon:damon_aggregated: \
            target_id=18446623435582458880 nr_regions=17 \
            140731667070976-140731668037632: 0 3

    Note that the last field is not in the early version[1].

    line is like below for damos_before_apply:

    kdamond.0 47293 [000] 80801.060214: damon:damos_before_apply: \
            ctx_idx=0 scheme_idx=0 target_idx=0 nr_regions=11 \
            121932607488-135128711168: 0 136

    [1] https://lore.kernel.org/linux-mm/df8d52f1fb2f353a62ff34dc09fe99e32ca1f63f.1636610337.git.xhao@linux.alibaba.com/
    '''

    if len(fields) != 12:
        return None, None, None, None

    end_time = int(float(fields[3][:-1]) * 1000000000)
    target_id = int(fields[7].split('=')[1])
    nr_regions = int(fields[8].split('=')[1])

    start_addr, end_addr = [int(x) for x in fields[9][:-1].split('-')]
    nr_accesses = int(fields[10])
    age = int(fields[11])
    region = _damon.DamonRegion(start_addr, end_addr, nr_accesses,
            _damon.unit_samples, age, _damon.unit_aggr_intervals)

    return region, end_time, target_id, nr_regions

def parse_perf_script_line(line):
    '''
    line could be that for damon_aggregated or damos_before_apply events
    '''
    fields = line.strip().split()
    if not len(fields) > 5:
        return None, None, None, None
    traceevent = fields[4][:-1]
    if traceevent == perf_event_damon_aggregated:
        return parse_damon_aggregated_perf_script_fields(fields)
    elif traceevent == perf_event_damos_before_apply:
        return parse_damos_before_apply_perf_script_fields(fields)
    else:
        return None, None, None, None

def parse_perf_script(script_output, monitoring_intervals):
    records = []
    snapshot = None

    for line in script_output.split('\n'):
        region, end_time, target_id, nr_regions = parse_perf_script_line(line)
        if region is None:
            continue

        record = record_of(target_id, records, monitoring_intervals)
        if len(record.snapshots) == 0:
            start_time = None
        else:
            start_time = record.snapshots[-1].end_time
            if start_time > end_time:
                return None, 'trace is not time-sorted'

        if snapshot is None:
            snapshot = DamonSnapshot(start_time, end_time, [], None)
            record.snapshots.append(snapshot)
        snapshot = record.snapshots[-1]
        snapshot.regions.append(region)

        if len(snapshot.regions) == nr_regions:
            snapshot = None

    for record in records:
        for snapshot in record.snapshots:
            snapshot.update_total_bytes()

    set_first_snapshot_start_time(records)
    return records, None

def set_perf_path(perf_path):
    global PERF
    PERF = perf_path

    # Test perf record for damon event
    err = None
    if _damo_subproc.avail_cmd(PERF):
        try:
            subprocess.check_output(
                    [PERF, 'record', '-e', perf_event_damon_aggregated, '--',
                        'sleep', '0'],
                    stderr=subprocess.PIPE)
        except:
            err = 'perf record not working with "%s"' % PERF
    else:
        err = 'perf not found at "%s"' % PERF
    return err

def parse_json(json_str):
    kvpairs = json.loads(json_str)
    return [DamonRecord.from_kvpairs(kvp) for kvp in kvpairs]

def parse_compressed_json(record_file):
    with open(record_file, 'rb') as f:
        compressed = f.read()
    decompressed = zlib.decompress(compressed).decode()
    return parse_json(decompressed)

def parse_json_file(record_file):
    with open(record_file) as f:
        json_str = f.read()
    return parse_json(json_str)

def parse_records_file(record_file, monitoring_intervals=None):
    '''
    Return monitoring results records and error string
    '''

    if not _damo_subproc.avail_cmd('file'):
        return None, "'file' command seems not installed"

    file_type = subprocess.check_output(
            ['file', '-b', record_file]).decode().strip()
    if file_type == 'JSON data':
        try:
            return parse_json_file(record_file), None
        except Exception as e:
            return None, 'failed parsing json file (%s)' % e
    if file_type == 'zlib compressed data':
        try:
            return parse_compressed_json(record_file), None
        except Exception as e:
            return None, 'failed parsing json compressed file (%s)' % e

    perf_script_output = None
    if file_type == 'ASCII text':
        with open(record_file, 'r') as f:
            perf_script_output = f.read()
    else:
        # might be perf data
        try:
            with open(os.devnull, 'w') as fnull:
                # in some setup, perf record file ends up having no proper
                # ownership.  There's no reason to be strict about that from
                # damo.  As long as we can, just parse it with '--force'
                # option.
                perf_script_output = subprocess.check_output(
                        [PERF, 'script', '--force', '-i', record_file],
                        stderr=fnull).decode()
        except:
            # Should be record format file
            pass
    if perf_script_output is not None:
        return parse_perf_script(perf_script_output, monitoring_intervals)
    else:
        return None, 'parsing %s failed' % record_file

# for writing monitoring results to a file

def write_json_compressed(records, file_path):
    json_str = json.dumps([r.to_kvpairs(raw=True) for r in records], indent=4)
    compressed = zlib.compress(json_str.encode())
    with open(file_path, 'wb') as f:
        f.write(compressed)

def write_json(records, file_path):
    json_str = json.dumps([r.to_kvpairs(raw=True) for r in records], indent=4)
    with open(file_path, 'w') as f:
        f.write(json_str)

def add_fake_snapshot_if_needed(records):
    '''
    perf and record file format stores only snapshot end time.  For a record
    having only single snapshot, hence, the reader of the files cannot know the
    start time of the snapshot.  Add a fake snapshot for the case.
    '''

    for record in records:
        snapshots = record.snapshots
        if len(snapshots) != 1:
            continue
        snapshot = snapshots[0]
        snap_duration = snapshot.end_time - snapshot.start_time
        # -1 nr_accesses.samples / -1 age.aggr_intervals means fake
        fake_regions = [_damon.DamonRegion(0, 0,
            -1, _damon.unit_samples, -1, _damon.unit_aggr_intervals)]
        fake_snapshot = DamonSnapshot(snapshot.end_time,
                snapshot.end_time + snap_duration, fake_regions, None)
        snapshots.append(fake_snapshot)

def write_perf_script(records, file_path):
    '''
    Example of the normal perf script output:

    kdamond.0  4452 [000] 82877.315633: damon:damon_aggregated: \
            target_id=18446623435582458880 nr_regions=17 \
            140731667070976-140731668037632: 0 3
    '''

    add_fake_snapshot_if_needed(records)
    with open(file_path, 'w') as f:
        for record in records:
            snapshots = record.snapshots
            for snapshot in snapshots:
                for region in snapshot.regions:
                    f.write(' '.join(['kdamond.x', 'xxxx', 'xxxx',
                        '%f:' % (snapshot.end_time / 1000000000.0),
                        'damon:damon_aggregated:',
                        'target_id=%s' % record.target_id,
                        'nr_regions=%d' % len(snapshot.regions),
                        '%d-%d: %d %s' % (region.start, region.end,
                            region.nr_accesses.samples,
                            region.age.aggr_intervals)]) + '\n')

def parse_file_permission_str(file_permission_str):
    try:
        file_permission = int(file_permission_str, 8)
    except Exception as e:
        return None, 'parsing failed (%s)' % e
    if file_permission < 0o0 or file_permission > 0o777:
        return None, 'out of available permission range'
    return file_permission, None

file_type_perf_script = 'perf_script'   # perf script output
file_type_perf_data = 'perf_data'       # perf record result file
file_type_json = 'json'                 # list of DamonRecord objects in json
file_type_json_compressed = 'json_compressed'

file_types = [file_type_json_compressed, file_type_json, file_type_perf_script,
        file_type_perf_data]
self_write_supported_file_types = [file_type_json_compressed, file_type_json,
        file_type_perf_script]

def write_damon_records(records, file_path, file_type, file_permission=None):
    '''Returns None if success, an error string otherwise'''
    if not file_type in self_write_supported_file_types:
        return 'write unsupported file type: %s' % file_type

    if file_type == file_type_json_compressed:
        write_json_compressed(records, file_path)
    elif file_type == file_type_json:
        write_json(records, file_path)
    elif file_type == file_type_perf_script:
        write_perf_script(records, file_path)

    if file_permission is not None:
        os.chmod(file_path, file_permission)
    return None

def rewrite_record_file(src_file, dst_file, file_format, file_permission=None,
        monitoring_intervals=None):
    records, err = parse_records_file(src_file, monitoring_intervals)
    if err:
        return err
    return write_damon_records(records, dst_file, file_format,
            file_permission)

def update_records_file(file_path, file_format, file_permission=None,
        monitoring_intervals=None):
    return rewrite_record_file(file_path, file_path, file_format,
            file_permission, monitoring_intervals)

# for recording

# Meaning of the fields of ProcMemFootprint are as below.
#
# ======== ===============================       ==============================
# Field    Content
# ======== ===============================       ==============================
# size     total program size (pages)            (same as VmSize in status)
# resident size of memory portions (pages)       (same as VmRSS in status)
# shared   number of pages that are shared       (i.e. backed by a file, same
#                                                as RssFile+RssShmem in status)
# trs      number of pages that are 'code'       (not including libs; broken,
#                                                includes data segment)
# lrs      number of pages of library            (always 0 on 2.6)
# drs      number of pages of data/stack         (including libs; broken,
#                                                includes library text)
# dt       number of dirty pages                 (always 0 on 2.6)
# ======== ===============================       ==============================
#
# The above table is stolen from Documentation/filesystems/proc.rst file of
# Linux
class ProcMemFootprint:
    size = None
    resident = None
    shared = None
    trs = None
    lrs = None
    drs = None
    dt = None

    def __init__(self, pid=None):
        if pid is None:
            return

        try:
            with open('/proc/%s/statm' % pid, 'r') as f:
                fields = [int(x) for x in f.read().split()]
        except:
            # the process may terminated.  Just think it as not using memory.
            fields = [0 for _ in range(7)]
        self.size = fields[0]
        self.resident = fields[1]
        self.shared = fields[2]
        self.trs = fields[3]
        self.lrs = fields[4]
        self.drs = fields[5]
        self.dt = fields[6]

    def to_kvpairs(self):
        return self.__dict__

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls()
        self.size = kvpairs['size']
        self.resident = kvpairs['resident']
        self.shared = kvpairs['shared']
        self.trs = kvpairs['trs']
        self.lrs = kvpairs['lrs']
        self.drs = kvpairs['drs']
        self.dt = kvpairs['dt']
        return self

class SysMemFootprint:
    total = None
    free = None
    available = None
    buffers = None
    cached = None

    def __init__(self, populate):
        if populate is False:
            return
        with open('/proc/meminfo', 'r') as f:
            for line in f:
                fields = line.split()
                if fields[0] == 'MemTotal:':
                    self.total = int(fields[1])
                if fields[0] == 'MemFree:':
                    self.free = int(fields[1])
                if fields[0] == 'MemAvailable:':
                    self.available = int(fields[1])
                if fields[0] == 'Buffers:':
                    self.buffers = int(fields[1])
                if fields[0] == 'Cached:':
                    self.cached = int(fields[1])

    def to_kvpairs(self):
        return self.__dict__

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(populate=False)
        self.total = kvpairs['total']
        self.free = kvpairs['free']
        self.available = kvpairs['available']
        self.buffers = kvpairs['buffers']
        self.cached = kvpairs['cached']
        return self

class MemFootprintsSnapshot:
    time = None
    footprints = None

    def __init__(self, pids=None):
        if pids is None:
            return

        self.time = time.time()
        self.footprints = {}
        for pid in pids:
            self.footprints[pid] = ProcMemFootprint(pid)
        self.footprints['sys'] = SysMemFootprint(populate=True)

    def to_kvpairs(self):
        footprints = []
        for target, fp in self.footprints.items():
            if target == 'sys':
                pid = None
            else:
                pid = target
            footprints.append({'pid': pid, 'footprint': fp.to_kvpairs()})
        return {'time': self.time, 'footprints': footprints}

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls()
        self.time = kvpairs['time']
        self.footprints = {}
        for fp in kvpairs['footprints']:
            pid, footprint = fp['pid'], fp['footprint']
            if pid is None:
                self.footprints[pid] = SysMemFootprint.from_kvpairs(footprint)
            else:
                self.footprints[pid] = ProcMemFootprint.from_kvpairs(footprint)
        return self

def record_mem_footprint(kdamonds, snapshots):
    pids = []
    for kdamond in kdamonds:
        for ctx in kdamond.contexts:
            for target in ctx.targets:
                if target.pid is None:
                    continue
                pids.append(target.pid)
    snapshots.append(MemFootprintsSnapshot(pids))

def save_mem_footprint(snapshots, filepath, file_permission):
    with open(filepath, 'w') as f:
        json.dump([s.to_kvpairs() for s in snapshots], f, indent=4)
    os.chmod(filepath, file_permission)

def load_mem_footprint(filepath):
    with open(filepath, 'r') as f:
        kvpairs = json.load(f)
    return [MemFootprintsSnapshot.from_kvpairs(x) for x in kvpairs]

class Vma:
    start = None
    end = None
    name = None

    def __init__(self, start, end, name):
        self.start = start
        self.end = end
        self.name = name

    def to_kvpairs(self):
        return self.__dict__

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(None, None, None)
        self.start = kvpairs['start']
        self.end = kvpairs['end']
        self.name = kvpairs['name']

class ProcVmas:
    pid = None
    vmas = None

    def __init__(self, pid):
        self.pid = pid
        self.vmas = []

        if pid is None:
            return

        try:
            with open('/proc/%s/maps' % pid, 'r') as f:
                for line in f:
                    fields = line.split()
                    start, end = [int(addr, 16) for addr in fields[0].split('-')]
                    name = fields[-1]
                    self.vmas.append(Vma(start, end, name))
        except:
            # the process may terminated.
            return

    def to_kvpairs(self):
        kvpairs = {'pid': self.pid}
        kvpairs['vmas'] = [v.to_kvpairs() for v in self.vmas]
        return kvpairs

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(None)
        self.pid = kvpairs['pid']
        self.vmas = [Vma.from_kvpairs(kvp) for kvp in kvpairs['vmas']]

class ProcVmasSnapshot:
    time = None
    procvmas = None

    def __init__(self, pids):
        if pids is None:
            return

        self.time = time.time()
        self.procvmas = []
        for pid in pids:
            self.procvmas.append(ProcVmas(pid))

    def to_kvpairs(self):
        kvpairs = {'time': self.time}
        kvpairs['procvmas'] = [p.to_kvpairs() for p in self.procvmas]
        return kvpairs

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(None)
        self.time = kvpairs['time']
        self.procvmas = [ProcVmas.from_kvpairs(kvp)
                         for kvp in kvapirs['procvmas']]

def record_proc_vmas(kdamonds, snapshots):
    pids = []
    for kdamond in kdamonds:
        for ctx in kdamond.contexts:
            for target in ctx.targets:
                if target.pid is None:
                    continue
                pids.append(target.pid)
    snapshots.append(ProcVmasSnapshot(pids))

def save_proc_vmas(snapshots, filepath, file_permission):
    with open(filepath, 'w') as f:
        json.dump([s.to_kvpairs() for s in snapshots], f, indent=4)
    os.chmod(filepath, file_permission)

def load_proc_vmas(filepath):
    with open(filepath, 'r') as f:
        kvpairs = json.load(f)
    return [ProcVmasSnapshot.from_kvpairs(x) for x in kvpairs]

class ProcStat:
    fields = None

    def __init__(self, pid):
        if pid is None:
            return
        try:
            with open('/proc/%s/stat' % pid, 'r') as f:
                self.fields = f.read().split()
        except Exception:
            # the process may finished already
            pass

    def to_kvpairs(self):
        return {
                'fields': self.fields,
                }

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(None)
        self.fields = kvpairs['fields']

    def __str__(self):
        if self.fields is None:
            return 'uninitialized'
        # from Documentation/filesystesm/proc.rst of Linux source tree
        field_names = ['pid', 'tcomm', 'state', 'ppid', 'pgrp', 'sid',
                       'tty_nr', 'tty_pgrp', 'flags', 'min_flt', 'cmin_flt',
                       'maj_flt', 'cmaj_flt', 'utime', 'stime', 'cutime',
                       'cstime', 'priority', 'nice', 'num_threads',
                       'it_real_value', 'start_time', 'vsize', 'rss', 'rsslim',
                       'start_code', 'end_code', 'start_stack', 'esp', 'eip',
                       'pending', 'blocked', 'sigign', 'sigcatch', '0', '0',
                       '0', 'exit_signal', 'task_cpu', 'rt_priority', 'policy',
                       'blkio_ticks', 'gtime', 'cgtime', 'start_data',
                       'end_data', 'start_brk', 'arg_start', 'arg_end',
                       'env_start', 'env_end', 'exit_code']
        lines = []
        for idx, field_name in enumerate(field_names):
            lines.append('%s: %s' % (field_name, self.fields[idx]))
        return '\n'.join(lines)

class ProcStatsSnapshot:
    time = None
    proc_stats = None

    def __init__(self, pids):
        if pids is None:
            return

        self.time = time.time()
        self.proc_stats = []
        for pid in pids:
            self.proc_stats.append(ProcStat(pid))

    def to_kvpairs(self):
        kvpairs = {'time': self.time}
        kvpairs['proc_stats'] = [p.to_kvpairs() for p in self.proc_stats]
        return kvpairs

    @classmethod
    def from_kvpairs(cls, kvpairs):
        self = cls(None)
        self.time = kvpairs['time']
        self.proc_stats = [ProcStat.from_kvpairs(kvp)
                         for kvp in kvapirs['proc_stats']]

def record_proc_stats(kdamonds, snapshots):
    pids = []
    for kdamond in kdamonds:
        pids.append(kdamond.pid)
        for ctx in kdamond.contexts:
            for target in ctx.targets:
                if target.pid is None:
                    continue
                pids.append(target.pid)
    snapshots.append(ProcStatsSnapshot(pids))

def save_proc_stats(snapshots, filepath, file_permission):
    with open(filepath, 'w') as f:
        json.dump([s.to_kvpairs() for s in snapshots], f, indent=4)
    os.chmod(filepath, file_permission)

def load_proc_stats(filepath):
    with open(filepath, 'r') as f:
        kvpairs = json.load(f)
    return [ProcStatsSnapshot.from_kvpairs(x) for x in kvpairs]

def add_childs_target(kdamonds):
    # TODO: Support multiple kdamonds
    if not _damon.target_has_pid(kdamonds[0].contexts[0].ops):
        return
    current_targets = kdamonds[0].contexts[0].targets

    for target in current_targets:
        if target.pid is None:
            continue
        try:
            childs_pids = subprocess.check_output(
                    ['ps', '--ppid', '%s' % target.pid, '-o', 'pid=']
                    ).decode().split()
        except:
            childs_pids = []
        if len(childs_pids) == 0:
            continue

        # TODO: Commit all at once, out of this loop
        new_targets = []
        for child_pid in childs_pids:
            # skip the child if already in the targets
            if child_pid in ['%s' % t.pid for t in current_targets]:
                continue
            # remove already terminated targets, since committing already
            # terminated targets to DAMON fails
            new_targets = [target for target in current_targets
                    if pid_running('%s' % target.pid)]
            new_targets.append(_damon.DamonTarget(pid=child_pid, regions=[]))
        if new_targets == []:
            continue

        # commit the new set of targets
        kdamonds[0].contexts[0].targets = new_targets
        err = _damon.commit(kdamonds)
        if err is not None:
            return 'commit failed (%s)' % err
    return None

def pid_running(pid):
    '''pid should be string'''
    try:
        subprocess.check_output(['ps', '--pid', pid])
        return True
    except:
        return False

def all_targets_terminated(targets):
    for target in targets:
        if pid_running('%s' % target.pid):
            return False
    return True

def poll_target_pids(kdamonds):
    '''Return True if >=1 target processes are running'''
    has_running_process = False
    if not kdamonds:
        return False
    for kdamond in kdamonds:
        for ctx in kdamond.contexts:
            if not _damon.target_has_pid(ctx.ops):
                continue
            if not all_targets_terminated(ctx.targets):
                return True
    return False

class RecordingHandle:
    '''
    Specifies the recording request.  The recording can be started by
    start_recording(), and finished by finish_recording().

    As a result of the two function calls, below files can be generated.

    - f'{handle.file_path}.kdamonds'
      - Given Kdamonds for the recording.  Have a json list of Kdamond kvpair
        objects.
    f'{handle.file_path}'
      - The DAMON monitoring results.  Have a json list of DamonRecord kvpair
        objects.  Each DamonRecord is for each target or scheme of the kdamond.
    f'{handle.file_path}.profile'
      - 'perf record' output file.
    f'{handle.file_path}.mem_footprint'
      - System and monitoring target process memory footprints.  Have a json
        list of MemFootprintsSnapshot kvpair objects.
    f'{handle.file_path}.vmas'
      - Monitoring target process memory mappings.  Have a json list of
        ProcVmaSnapshot kvpair objects.
    f'{handle.file_path}.proc_stats'
      - Kdamonds and monitoring target processes /proc/PID/stat contents.  Have
        a json list of ProcStatsSnapshot kvpair objects.
    '''

    file_path = None
    file_format = None
    file_permission = None

    # for access patterns tracing
    tracepoint = None
    monitoring_intervals = None
    perf_pipe = None

    # for access patterns snapshot
    snapshot_request = None
    snapshot_records = None
    snapshot_count = None
    snapshot_interval_sec = None

    # for CPU clock event recording
    do_profile = None
    perf_profile_pipe = None

    # for adding child tasks and memory footprint recording
    kdamonds = None
    add_child_tasks = None
    mem_footprint_snapshots = None

    # for vmas recording
    vmas_snapshots = None

    # for /proc/<pid>/stat recording
    proc_stats = None

    timeout = None

    def __init__(self, tracepoint, file_path, file_format, file_permission,
                 monitoring_intervals,
                 do_profile,
                 kdamonds, add_child_tasks, record_mem_footprint,
                 record_vmas, record_proc_stats, timeout, snapshot_request,
                 snapshot_interval_sec, snapshot_count):
        self.tracepoint = tracepoint
        self.file_path = file_path
        self.file_format = file_format
        self.file_permission = file_permission
        self.monitoring_intervals = monitoring_intervals

        self.do_profile = do_profile

        self.kdamonds = kdamonds
        self.add_child_tasks = add_child_tasks
        if record_mem_footprint is True:
            self.mem_footprint_snapshots = []
        if record_vmas is True:
            self.vmas_snapshots = []

        if record_proc_stats is True:
            self.proc_stats = []

        self.timeout = timeout

        self.snapshot_request = snapshot_request
        self.snapshot_interval_sec = snapshot_interval_sec
        self.snapshot_count = snapshot_count

    def will_take_awhile(self):
        if self.snapshot_interval_sec == 0 and self.snapshot_count < 5:
            return False
        return True

def start_recording(handle):
    kdamonds_file_path = '%s.kdamonds' % handle.file_path
    with open(kdamonds_file_path, 'w') as f:
        json.dump([k.to_kvpairs() for k in handle.kdamonds], f, indent=4)
    os.chmod(kdamonds_file_path, handle.file_permission)

    if handle.tracepoint is not None:
        handle.perf_pipe = subprocess.Popen(
                [PERF, 'record', '-a', '-e', handle.tracepoint,
                 '-o', handle.file_path])
    if handle.do_profile:
        cmd = [PERF, 'record', '-o', '%s.profile' % handle.file_path]
        handle.perf_profile_pipe = subprocess.Popen(cmd)

    start_time = time.time()
    nr_snapshots_to_take = handle.snapshot_count
    if handle.snapshot_interval_sec:
        sleep_time = handle.snapshot_interval_sec
    else:
        sleep_time = 1
    while (poll_target_pids(handle.kdamonds) or
           _damon.any_kdamond_running()):
        if handle.add_child_tasks is True:
            add_childs_target(handle.kdamonds)

        if handle.mem_footprint_snapshots is not None:
            record_mem_footprint(handle.kdamonds,
                                 handle.mem_footprint_snapshots)
        if handle.vmas_snapshots is not None:
            record_proc_vmas(handle.kdamonds, handle.vmas_snapshots)

        if handle.proc_stats is not None:
            record_proc_stats(handle.kdamonds, handle.proc_stats)

        if (handle.timeout is not None and
            time.time() - start_time >= handle.timeout):
            break

        if handle.snapshot_request:
            if handle.snapshot_records is None:
                handle.snapshot_records = []
            snapshot_records, err = get_snapshot_records_of(
                    handle.snapshot_request)
            if err is not None:
                print('failed getting snapshot')
                exit(1)
            handle.snapshot_records += snapshot_records
            nr_snapshots_to_take -= 1
            if nr_snapshots_to_take == 0:
                handle.snapshot_records = merge_records(
                        handle.snapshot_records)
                break

        time.sleep(sleep_time)

def finish_recording(handle):
    if handle.perf_pipe:
        try:
            handle.perf_pipe.send_signal(signal.SIGINT)
            handle.perf_pipe.wait()
        except:
            # perf might already finished
            pass

        if handle.file_format == file_type_perf_data:
            os.chmod(handle.file_path, handle.file_permission)
        else:
            err = update_records_file(handle.file_path, handle.file_format,
                    handle.file_permission, handle.monitoring_intervals)
            if err is not None:
                print('converting format from perf_data to %s failed (%s)' %
                        (handle.file_format, err))

    if handle.snapshot_records:
        write_damon_records(handle.snapshot_records, handle.file_path,
                            handle.file_format, handle.file_permission)

    if handle.perf_profile_pipe is not None:
        try:
            handle.perf_profile_pipe.send_signal(signal.SIGINT)
        except:
            # perf might already finished
            pass
        os.chmod('%s.profile' % handle.file_path, handle.file_permission)

    if handle.mem_footprint_snapshots is not None:
        save_mem_footprint(
                handle.mem_footprint_snapshots,
                '%s.mem_footprint' % handle.file_path, handle.file_permission)
    if handle.vmas_snapshots is not None:
        save_proc_vmas(handle.vmas_snapshots, '%s.vmas' % handle.file_path,
                       handle.file_permission)
    if handle.proc_stats is not None:
        save_proc_stats(handle.proc_stats, '%s.proc_stats' % handle.file_path,
                        handle.file_permission)

# for snapshot

def current_kdamonds_interval_updated():
    kdamonds = _damon.current_kdamonds()
    intervals_autotuning = False
    for kd in kdamonds:
        for ctx in kd.contexts:
            if ctx.intervals.intervals_goal.enabled():
                intervals_autotuning = True
                break
        if intervals_autotuning:
            break
    if not intervals_autotuning:
        return kdamonds
    _damon.update_tuned_intervals()
    return _damon.current_kdamonds()

def find_install_scheme(scheme_to_find):
    '''Install given scheme to all contexts if effectively same scheme is not
    installed.
    Returns whether it found a context doesn't having the scheme, indices list
    for the effectively same schemes, and an error if something wrong.
    '''
    installed = False
    indices = []
    kdamonds = current_kdamonds_interval_updated()
    for kidx, kdamond in enumerate(kdamonds):
        for cidx, ctx in enumerate(kdamond.contexts):
            ctx_has_the_scheme = False
            for sidx, scheme in enumerate(ctx.schemes):
                if scheme.effectively_equal(scheme_to_find, ctx.intervals):
                    if _damon.feature_supported('schemes_apply_interval'):
                        scheme_to_find.apply_interval_us = ctx.intervals.sample
                    ctx_has_the_scheme = True
                    indices.append([kidx, cidx, sidx])
                    break
            if ctx_has_the_scheme:
                continue

            if _damon.feature_supported('schemes_apply_interval'):
                scheme_to_find.apply_interval_us = ctx.intervals.sample
            ctx.schemes.append(scheme_to_find)
            installed = True
            indices.append([kidx, cidx, len(ctx.schemes) - 1])
    if installed:
        err = _damon.commit(kdamonds)
        if err is not None:
            return (False, [], None,
                    'committing scheme installed kdamonds failed: %s' % err)
    return installed, indices, kdamonds, None

def can_merge(left_region, right_region):
    if left_region.sz_filter_passed or right_region.sz_filter_passed:
        return False
    return (left_region.end == right_region.start and
            left_region.nr_accesses == right_region.nr_accesses and
            left_region.age == right_region.age)

def tried_regions_to_snapshot(scheme, intervals, merge_regions):
    snapshot_end_time_ns = time.time() * 1000000000
    snapshot_start_time_ns = snapshot_end_time_ns - intervals.aggr * 1000
    regions = []

    for tried_region in scheme.tried_regions:
        '''Merge regions that having same access pattern, since DAMON usually
        splits regions unnecessarily to keep the min_nr_regions'''
        if merge_regions and len(regions) > 0:
            last_region = regions[-1]
            if can_merge(last_region, tried_region):
                last_region.end = tried_region.end
                if last_region.sz_filter_passed is not None:
                    last_region.sz_filter_passed += tried_region.sz_filter_passed
                continue
        regions.append(tried_region)
    if scheme.tried_bytes is not None:
        total_bytes = scheme.tried_bytes
    else:
        total_bytes = None

    return DamonSnapshot(snapshot_start_time_ns, snapshot_end_time_ns, regions,
            total_bytes, scheme.stats)

def tried_regions_to_records_of(idxs, merge_regions):
    '''idxs: list of kdamond/context/scheme indices to get records for.  If it
    is None, return records for all schemes'''
    records = []
    for kdamond_idx, kdamond in enumerate(_damon.current_kdamonds()):
        if kdamond.state != 'on':
            continue
        for ctx_idx, ctx in enumerate(kdamond.contexts):
            for scheme_idx, scheme in enumerate(ctx.schemes):
                if not [kdamond_idx, ctx_idx, scheme_idx] in idxs:
                    continue

                snapshot = tried_regions_to_snapshot(scheme, ctx.intervals,
                        merge_regions)

                records.append(DamonRecord(kdamond_idx, ctx_idx, ctx.intervals,
                    scheme_idx, target_id=None, scheme_filters=scheme.filters))
                records[-1].snapshots.append(snapshot)
                break
    return records

def three_regions_of(pid):
    '''
    Return three big mapped virtual address ranges of a given process, which
    separated by the two huge gaps[1].

    [1] https://docs.kernel.org/mm/damon/design.html#vma-based-target-address-range-construction
    '''
    if not os.path.isfile('/proc/%s/maps' % pid):
        print('maps file for %s pid not found' % pid)
        exit(0)
    with open('/proc/%s/maps' % pid, 'r') as f:
        maps_content = f.read()
    regions = []
    for line in maps_content.split('\n'):
        if line == '':
            continue
        start, end = [int(addr, 16) for addr in line.split()[0].split('-')]
        if len(regions) > 0 and regions[-1].end == start:
            regions[-1].end = end
        else:
            regions.append(_damon.DamonRegion(start, end))

    gaps = []
    for idx, region in enumerate(regions):
        if idx == 0:
            continue
        prev_region = regions[idx - 1]
        if region.start != prev_region.end:
            gaps.append([prev_region.end, region.start])
    gaps.sort(key=lambda x: x[1] - x[0], reverse=True)
    if len(gaps) < 2:
        return regions
    # sort biggest two gaps in address
    gaps = sorted(gaps[:2], key=lambda x: x[0])

    return [_damon.DamonRegion(regions[0].start, gaps[0][0]),
            _damon.DamonRegion(gaps[0][1], gaps[1][0]),
            _damon.DamonRegion(gaps[1][1], regions[-1].end)]

def install_target_regions_if_needed(kdamonds):
    '''Returns an error string, or None'''
    need_install = False
    for kd in kdamonds:
        for ctx in kd.contexts:
            if ctx.ops != 'vaddr':
                continue
            need_install = True
            for target in ctx.targets:
                target.regions = three_regions_of(target.pid)
    if not need_install:
        return None
    err = _damon.commit(kdamonds)
    for kd in kdamonds:
        for ctx in kd.contexts:
            if ctx.ops != 'vaddr':
                continue
            for target in ctx.targets:
                target.regions = []
    return err

def update_get_snapshot_records(kdamond_idxs, scheme_idxs,
        total_sz_only, merge_regions):
    if total_sz_only:
        err = _damon.update_schemes_tried_bytes(kdamond_idxs)
        # update_schemes_tried_bytes() can error if the feature is not
        # supported.  Then, full record will be returned
        if err is None:
            records = tried_regions_to_records_of(scheme_idxs, merge_regions)
            return records, None

    err = 'assumed error'
    nr_tries = 0
    while err is not None and nr_tries < 5:
        nr_tries += 1

        # todo: update schemes stats only if really required
        err = _damon.update_schemes_stats(kdamond_idxs)
        if err is not None:
            time.sleep(random.rndrange(
                2**(nr_tries - 1), 2**nr_tries) / 100)
            continue

        # todo: update tuned intervals only if auto-tuning is ongoing
        err = _damon.update_tuned_intervals(kdamond_idxs)
        if err is not None:
            time.sleep(random.randrange(
                2**(nr_tries - 1), 2**nr_tries) / 100)
            continue

        err = _damon.update_schemes_tried_regions(kdamond_idxs)

        if err is not None:
            time.sleep(random.randrange(
                2**(nr_tries - 1), 2**nr_tries) / 100)
    if err is not None:
        return None, 'updating schemes tried regions fail: %s' % err

    records = tried_regions_to_records_of(scheme_idxs, merge_regions)
    return records, None

def kdamonds_copy_intervals(src_kdamonds, dst_kdamonds):
    for kidx, kdamond in enumerate(src_kdamonds):
        if kidx >= len(dst_kdamonds):
            continue
        dst_kdamond = dst_kdamonds[kidx]
        for cidx, ctx in enumerate(kdamond.contexts):
            if cidx >= len(dst_kdamond.contexts):
                continue
            dst_ctx = dst_kdamond.contexts[cidx]
            dst_ctx.intervals = ctx.intervals

def get_snapshot_records(monitor_scheme, total_sz_only, merge_regions):
    'return DamonRecord objects each having single DamonSnapshot and an error'
    running_kdamond_idxs = _damon.running_kdamond_idxs()
    if len(running_kdamond_idxs) == 0:
        return None, 'no kdamond running'

    orig_kdamonds = _damon.current_kdamonds()

    err = install_target_regions_if_needed(orig_kdamonds)
    if err is not None:
        return None, 'vaddr region install failed (%s)' % err

    installed, idxs, updated_kdamonds, err = find_install_scheme(
            monitor_scheme)
    if err:
        return None, 'monitoring scheme install failed: %s' % err

    records, err = update_get_snapshot_records(running_kdamond_idxs, idxs,
            total_sz_only, merge_regions)

    if installed:
        kdamonds_copy_intervals(updated_kdamonds, orig_kdamonds)
        uninstall_err = _damon.commit(orig_kdamonds)
        if uninstall_err:
            errmsg = 'monitoring scheme uninstall failed: %s' % uninstall_err
            if err is not None:
                err += ', %s' % errmsg
            else:
                err = errmsg

    return records, err

def get_snapshot_records_for_schemes(idxs, total_sz_only, merge_regions):
    '''idxs: list of kdamond/context/scheme indices to get records for.
    Return DamonRecord objects each having single DamonSnapshot and an error'''
    running_kdamond_idxs = _damon.running_kdamond_idxs()
    if len(running_kdamond_idxs) == 0:
        return None, 'no kdamond running'

    return update_get_snapshot_records(running_kdamond_idxs, idxs,
            total_sz_only, merge_regions)

def region_of_pattern(region, pattern, record_intervals):
    sz_bytes = pattern.sz_bytes
    nr_acc = pattern.nr_acc_min_max
    age = pattern.age_min_max
    # return if the region fits into the pattern
    sz = region.size()
    if sz < sz_bytes[0] or sz_bytes[1] < sz:
        return False

    if record_intervals is None:
        return True
    region.nr_accesses.add_unset_unit(record_intervals)
    freq = region.nr_accesses.percent
    if freq < nr_acc[0].percent or nr_acc[1].percent < freq:
        return False
    region.age.add_unset_unit(record_intervals)
    usecs = region.age.usec
    if usecs < age[0].usec or age[1].usec < usecs:
        return False
    return True

def filter_by_pattern(record, access_pattern):
    for snapshot in record.snapshots:
        filtered = []
        for region in snapshot.regions:
            if region_of_pattern(region, access_pattern, record.intervals):
                filtered.append(region)
        snapshot.regions = filtered

def filter_by_addr(region, addr_ranges):
    regions = []
    for start, end in addr_ranges:
        # out of the range
        if region.end <= start or end <= region.start:
            continue
        # in the range
        if start <= region.start and region.end <= end:
            regions.append(copy.deepcopy(region))
            continue
        # overlap
        copied = copy.deepcopy(region)
        copied.start = max(start, region.start)
        copied.end = min(end, region.end)
        regions.append(copied)
    return regions

def filter_records_by_addr(records, addr_ranges):
    for record in records:
        for snapshot in record.snapshots:
            filtered_regions = []
            for region in snapshot.regions:
                filtered_regions += filter_by_addr(region, addr_ranges)
            snapshot.regions = filtered_regions
            snapshot.update_total_bytes()

def filter_records_by_snapshot_sz(records, sz_ranges):
    for record in records:
        filtered_snapshots = []
        for snapshot in record.snapshots:
            for min_, max_ in sz_ranges:
                if (min_ <= snapshot.total_bytes and
                        snapshot.total_bytes <= max_):
                    filtered_snapshots.append(snapshot)
                    break
        record.snapshots = filtered_snapshots

def filter_records_by_snapshot_time(records, time_ranges):
    for record in records:
        filtered_snapshots = []
        for snapshot in record.snapshots:
            filter_out = True
            for start_sec, end_sec in time_ranges:
                if (snapshot.start_time >= start_sec and
                    snapshot.end_time <= end_sec):
                    filter_out = False
                    break
            if filter_out is False:
                filtered_snapshots.append(snapshot)
        record.snapshots = filtered_snapshots

def filter_records_by_snapshot_indices(records, index_ranges):
    for record in records:
        filtered_snapshots = []
        for idx, snapshot in enumerate(record.snapshots):
            filter_out = True
            for start_idx, end_idx in index_ranges:
                if start_idx <= idx and idx <= end_idx:
                    filter_out = False
                    break
            if filter_out is False:
                filtered_snapshots.append(snapshot)
        record.snapshots = filtered_snapshots

def filter_records_by_temperature(records, temperature_ranges,
                                  temperature_weights):
    for record in records:
        for snapshot in record.snapshots:
            filtered_regions = []
            for region in snapshot.regions:
                region.nr_accesses.add_unset_unit(record.intervals)
                region.age.add_unset_unit(record.intervals)
                temperature = damo_report_access.temperature_of(
                        region, temperature_weights)
                for min_t, max_t in temperature_ranges:
                    if min_t <= temperature and temperature <= max_t:
                        filtered_regions.append(region)
                        break
            snapshot.regions = filtered_regions
            snapshot.update_total_bytes()

def get_snapshot_records_of(request):
    '''
    get records containing single snapshot from running kdamonds
    '''
    if request.tried_regions_of is None:
        access_pattern = _damon.DamosAccessPattern()
        filters = []
        if request.record_filter:
            access_pattern = request.record_filter.access_pattern

            addr_ranges = request.record_filter.address_ranges
            if addr_ranges and _damon.feature_supported('schemes_filters_addr'):
                for start, end in addr_ranges:
                    filters.append(_damon.DamosFilter(
                        'addr', False,
                        address_range=_damon.DamonRegion(start, end)))
        if request.snapshot_damos_filters is not None:
            filters += request.snapshot_damos_filters
        monitor_scheme = _damon.Damos(
                access_pattern=access_pattern, filters=filters)

        records, err = get_snapshot_records(monitor_scheme,
                request.total_sz_only, not request.dont_merge_regions)
    else:
         records, err = get_snapshot_records_for_schemes(
                 request.tried_regions_of, request.total_sz_only,
                 not request.dont_merge_regions)
    return records, err

class RecordFilter:
    access_pattern = None
    address_ranges = None
    snapshot_sz_ranges = None
    snapshot_time_ranges = None
    snapshot_index_ranges = None
    temperature_ranges = None
    temperature_weights = None

    def __init__(self, access_pattern, address_ranges, snapshot_sz_ranges,
                 snapshot_time_ranges, snapshot_index_ranges,
                 temperature_ranges, temperature_weights):
        self.access_pattern = access_pattern
        self.address_ranges = address_ranges
        self.snapshot_sz_ranges = snapshot_sz_ranges
        self.snapshot_time_ranges = snapshot_time_ranges
        self.snapshot_index_ranges = snapshot_index_ranges
        self.temperature_ranges = temperature_ranges
        self.temperature_weights = temperature_weights

    def to_kvpairs(self, raw):
        kvpairs = {'access_pattern': self.access_pattern.to_kvpairs(raw)}
        if self.address_ranges is None:
            address_ranges = None
        else:
            address_ranges = [
                    [_damo_fmt_str.format_sz(start, raw),
                     _damo_fmt_str.format_sz(end, raw)]
                    for start, end in self.address_ranges]
        kvpairs['address_ranges'] = address_ranges
        if self.snapshot_sz_ranges is None:
            snapshot_sz_ranges = None
        else:
            snapshot_sz_ranges = [
                    [_damo_fmt_str.format_sz(start, raw),
                     _damo_fmt_str.format_sz(end, raw)]
                    for start, end in self.snapshot_sz_ranges]
        kvpairs['snapshot_sz_ranges'] = snapshot_sz_ranges
        kvpairs['snapshot_time_ranges'] = self.snapshot_time_ranges
        kvpairs['snapshot_index_ranges'] = self.snapshot_index_ranges
        kvpairs['temperature_ranges'] = self.temperature_ranges
        kvpairs['temperature_weights'] = self.temperature_weights
        return kvpairs

    def filter_records(self, records):
        if self.access_pattern is not None:
            for record in records:
                filter_by_pattern(record, self.access_pattern)
        if self.address_ranges is not None:
            filter_records_by_addr(records, self.address_ranges)
        if self.snapshot_sz_ranges is not None:
            filter_records_by_snapshot_sz(records, self.snapshot_sz_ranges)
        if self.snapshot_time_ranges is not None:
            filter_records_by_snapshot_time(records, self.snapshot_time_ranges)
        if self.snapshot_index_ranges is not None:
            filter_records_by_snapshot_indices(records,
                                               self.snapshot_index_ranges)
        if self.temperature_ranges is not None:
            filter_records_by_temperature(records, self.temperature_ranges,
                                          self.temperature_weights)

def set_snapshot_damos_filters_option(parser):
    parser.add_argument(
            '--snapshot_damos_filter', nargs='+', action='append', default=[],
            metavar='<damos filter argument>',
            help=' '.join([
                'Region-internal DAMOS filters for the snapshot.',
                'Format is same to --damos_filter.'
                ]))

class RecordGetRequest:
    # TODO: Extend to be used for recording

    # source of the record.  If both are None, get snapshot
    tried_regions_of = None
    record_file = None

    snapshot_damos_filters = None

    record_filter = None

    # more detailed requests
    total_sz_only = None
    dont_merge_regions = None

    def __init__(
            self, tried_regions_of=None, record_file=None,
            snapshot_damos_filters=None,
            record_filter=None,
            total_sz_only=False, dont_merge_regions=True):
        self.tried_regions_of = tried_regions_of
        self.record_file = record_file
        self.snapshot_damos_filters = snapshot_damos_filters
        self.record_filter = record_filter
        self.total_sz_only = total_sz_only
        self.dont_merge_regions = dont_merge_regions

def get_records(tried_regions_of=None, record_file=None,
                snapshot_damos_filters=None, record_filter=None,
                total_sz_only=False, dont_merge_regions=True):
    request = RecordGetRequest(
            tried_regions_of, record_file, snapshot_damos_filters,
            record_filter, total_sz_only, dont_merge_regions)

    # If record is live snapshot, access pattern filtering is applied with
    # get_snapshot_records_of() because it uses DAMOS to get the snapshot.  If
    # the kernel has schemes_filters_addr feature, address ranges filter is
    # also applied.  Also, snapshot time range filter makes no sense for live
    # snapshot.
    # To avoid confusing caller, copy filter, modify the filter as necessary
    # and apply only necessary filters at once at last stage.
    if record_filter:
        filter_copy = copy.deepcopy(record_filter)
    else:
        filter_copy = RecordFilter(None, None, None, None, None, None, None)

    if request.record_file is None:
        records, err = get_snapshot_records_of(request)
        if err is not None:
            return None, err
        filter_copy.access_pattern = None
        if _damon.feature_supported('schemes_filters_addr'):
            filter_copy.address_ranges = None
            filter_copy.snapshot_time_ranges = None
    else:
        if not os.path.isfile(request.record_file):
            return None, '%s not found' % request.record_file

        records, err = parse_records_file(request.record_file)
        if err:
            return None, ('parsing %s failed (%s)' %
                    (request.record_file, err))

    filter_copy.filter_records(records)
    return records, None

def parse_sort_bytes_ranges_input(bytes_ranges_input):
    try:
        ranges = [[_damo_fmt_str.text_to_bytes(start),
            _damo_fmt_str.text_to_bytes(end)]
            for start, end in bytes_ranges_input]
    except Exception as e:
        return None, 'conversion to bytes failed (%s)' % e

    ranges.sort(key=lambda x: x[0])
    for idx, arange in enumerate(ranges):
        start, end = arange
        if start > end:
            return None, 'start > end (%s)' % arange
        if idx > 0 and ranges[idx - 1][1] > start:
            return None, 'overlapping range'
    return ranges, None

def args_to_filter(args):
    access_pattern = _damon.DamosAccessPattern(args.sz_region,
            args.access_rate, _damon.unit_percent, args.age * 1000000,
            _damon.unit_usec)

    addr_range = None
    if args.address is not None:
        addr_range, err = parse_sort_bytes_ranges_input(
                args.address)
        if err is not None:
            return None, 'wrong --address input (%s)' % err

    snapshot_sz_ranges = None
    if args.sz_snapshot is not None:
        snapshot_sz_ranges, err = parse_sort_bytes_ranges_input(
                args.sz_snapshot)
        if err is not None:
            return None, 'wrong --sz_snapshot input (%s)' % err

    snapshot_time = None
    if args.snapshot_time is not None:
        snapshot_time = [
                [_damo_fmt_str.text_to_ns(s), _damo_fmt_str.text_to_ns(e)]
                for s, e in args.snapshot_time]

    snapshot_index_ranges = None
    if args.snapshot_idx is not None:
        snapshot_index_ranges = [
                [_damo_fmt_str.text_to_nr(s), _damo_fmt_str.text_to_nr(e)]
                for s, e in args.snapshot_idx]

    # maybe this is called from damo_report_access
    if hasattr(args, 'temperature_weights'):
        temperature_weights = args.temperature_weights
    else:
        # ignore size
        temperature_weights = [0, 100, 100]

    return RecordFilter(access_pattern, addr_range,
                        snapshot_sz_ranges, snapshot_time,
                        snapshot_index_ranges,
                        args.temperature, temperature_weights), None

def set_filter_argparser(parser, hide_help=False):
    parser.add_argument('--sz_region', metavar=('<min>', '<max>'), nargs=2,
                        default=['min', 'max'],
                        help='min/max size of regions (bytes) to filter in'
                        if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--access_rate', metavar=('<min>', '<max>'), nargs=2,
            default=['min', 'max'],
            help='min/max access rate of regions (percent) to filter in'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--age', metavar=('<min>', '<max>'), nargs=2,
            default=['min', 'max'],
            help='min/max age of regions (seconds) to filter in'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--address', metavar=('<start>', '<end>'), nargs=2,
            action='append',
            help='address ranges to filter in'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--sz_snapshot', metavar=('<min>', '<max>'), nargs=2,
            action='append',
            help='min/max total size of regions of snapshots to filter in'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--snapshot_time', metavar=('<start (ns)>', '<end (ns)>'), nargs=2,
            action='append',
            help='show snapshots generated in these time intervals'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--snapshot_idx', metavar=('<start>', '<end>'), nargs=2,
            action='append',
            help='show snapshots of these indices range'
            if not hide_help else argparse.SUPPRESS)
    parser.add_argument(
            '--temperature', metavar=('<min>', '<max>'), nargs=2, type=int,
            action='append',
            help='min/max temperature of regions to filter in'
            if not hide_help else argparse.SUPPRESS)
    if hide_help is True:
        parser.epilog += ' '.join([
            " More options for filtering are available.",
            "Do 'damo help access_filter_options -h' for those options.",
            ])



# SPDX-License-Identifier: GPL-2.0

import argparse
import subprocess

import _damo_dist
import _damo_fmt_str
import _damo_print
import _damo_records
import damo_record_info
import damo_report_footprint
import damo_report_heatmap
import damo_wss

def fmt_report_short(args):
    lines = []
    records, err = _damo_records.get_records(record_file=args.access_pattern)
    if err is not None:
        print('access pattern record file (%s) parsing failed (%s)' %
              (args.access_pattern, err))
        exit(1)

    guides = damo_record_info.get_guide_info(records)
    lines.append('# Heatmap')
    for guide in guides:
        for region in guide.regions():
            lines.append('# target %d, address range %d-%d' % (
                guide.tid, region[0], region[1]))
            heatmap = damo_report_heatmap.fmt_heats(
                    argparse.Namespace(
                        tid=guide.tid, resol=[5, 80],
                        time_range=[guide.start_time, guide.end_time],
                        address_range=region,
                        output='stdout',
                        stdout_colorset='gray',
                        stdout_skip_colorset_example=True,
                        ),
                    records)
            lines.append(heatmap)

    lines.append('')
    lines.append('# Memory Footprints Distribution')

    nr_snapshots = 0
    for record in records:
        nr_snapshots += len(record.snapshots)
    if nr_snapshots > 200:
        _damo_records.adjust_records(
                records, aggregate_interval=1, nr_snapshots_to_skip=20)

    if args.footprints is None:
        args.footprints = args.access_pattern + '.mem_footprint'

    lines.append('%10s %15s %15s %15s %15s %15s' %
                 ('percentile', '0', '25', '50', '75', '100'))
    wss_dists = damo_wss.get_wss_dists(
            records, acc_thres=1, sz_thres=1, do_sort=True,
            collapse_targets=True)
    for tid, dists in wss_dists.items():
        # because collapsed targets, only one iteration will be executed here
        line = '%10s ' % 'wss'
        for percentile in range(0, 101, 25):
            val = _damo_fmt_str.format_sz(
                    _damo_dist.get_percentile(dists, percentile), False)
            line += '%15s ' % val
        lines.append(line)
    for metric in ['rss', 'vsz', 'sys_used']:
        line = '%10s ' % metric
        fp_dists = damo_report_footprint.get_dists(
                records=args.footprints, metric=metric,
                do_sort=True)
        for percentile in range(0, 101, 25):
            val = _damo_fmt_str.format_sz(
                    _damo_dist.get_percentile(fp_dists, percentile), False)
            line += '%15s ' % val
        lines.append(line)

    lines.append('')
    lines.append('# Hotspot functions')
    if args.profile is None:
        args.profile = args.access_pattern + '.profile'

    cmd = [args.perf_path, 'report', '-i', args.profile, '--stdio']
    output_lines = subprocess.check_output(cmd).decode().split('\n')
    output_lines = output_lines[5:21]
    lines += output_lines

    return '\n'.join(lines)

def fmt_report(args):
    lines = []
    records, err = _damo_records.get_records(record_file=args.access_pattern)
    if err is not None:
        print('access pattern record file (%s) parsing failed (%s)' %
              (args.access_pattern, err))
        exit(1)

    guides = damo_record_info.get_guide_info(records)

    lines.append('Overall recorded access pattern')
    lines.append('===============================')
    lines.append('')
    for guide in guides:
        lines.append('%s' % guide)
    lines.append('')
    lines.append('# regions above are regions that access pattern recorded')
    lines.append('# you can get this via \'damo record_info\', too')

    lines.append('')
    lines.append('Heatmap')
    lines.append('=======')
    lines.append('')
    for guide in guides:
        lines.append('# target %d' % guide.tid)
        for region in guide.regions():
            lines.append('# address range %d-%d' % (region[0], region[1]))
            heatmap = damo_report_heatmap.fmt_heats(
                    argparse.Namespace(
                        tid=guide.tid, resol=[10, 80],
                        time_range=[guide.start_time, guide.end_time],
                        address_range=region,
                        output='stdout',
                        stdout_colorset='gray',
                        stdout_skip_colorset_example=True,
                        ),
                    records)
            lines.append(heatmap)
    lines.append('# you can get above via \'damo report heatmap\'')

    lines.append('')
    lines.append('Working Set Size and Memory Footprints Distribution')
    lines.append('===================================================')
    lines.append('')

    nr_snapshots = 0
    for record in records:
        nr_snapshots += len(record.snapshots)
    if nr_snapshots > 200:
        _damo_records.adjust_records(
                records, aggregate_interval=1, nr_snapshots_to_skip=20)

    if args.footprints is None:
        args.footprints = args.access_pattern + '.mem_footprint'

    for sort_key in ['size', 'time']:
        lines.append('Sorted by %s' % sort_key)
        lines.append('--------------')
        lines.append('')
        wss_dists = damo_wss.get_wss_dists(
                records, acc_thres=1, sz_thres=1, do_sort=sort_key == 'size',
                collapse_targets=True)
        for tid, dists in wss_dists.items():
            # because collapsed targets, only one iteration will be executed here
            output = _damo_dist.fmt_dists(
                    'wss', dists, range(0, 101, 25), pr_all=False,
                    format_fn=_damo_fmt_str.format_sz, raw_number=False,
                    nr_cols_bar=59)
            lines.append(output)
        for metric in ['rss', 'vsz', 'sys_used']:
            fp_dists = damo_report_footprint.get_dists(
                    records=args.footprints, metric=metric,
                    do_sort=sort_key == 'size')
            output = _damo_dist.fmt_dists(
                    metric, fp_dists, range(0, 101, 25), pr_all=False,
                    format_fn=_damo_fmt_str.format_sz, raw_number=False,
                    nr_cols_bar=59)
            lines.append(output)
        lines.append('')
    lines.append('# you can get above via \'damo report wss\' and \'damo report footprints\'')

    lines.append('')
    lines.append('Hotspot functions')
    lines.append('=================')
    lines.append('')

    if args.profile is None:
        args.profile = args.access_pattern + '.profile'

    cmd = [args.perf_path, 'report', '-i', args.profile, '--stdio']
    output = subprocess.check_output(cmd).decode()
    lines.append('\n'.join(output.split('\n')[:30]))

    lines.append('# you can get above via \'damo report profile\'')

    return '\n'.join(lines)

def main(args):
    if args.long:
        report_text = fmt_report(args)
        _damo_print.pr_with_pager_if_needed(report_text)
    else:
        report_text = fmt_report_short(args)
        print(report_text)

def set_argparser(parser):
    parser.add_argument(
            '--access_pattern', metavar='<file>', default='damon.data',
            help='access pattern record file')
    parser.add_argument(
            '--footprints', metavar='<file>',
            help='memory footprints record file')
    parser.add_argument(
            '--profile', metavar='<file>', help='profile record file')
    parser.add_argument(
            '--long', action='store_true', help='make long report')
    parser.add_argument('--perf_path', type=str, default='perf',
                        help='path of perf tool')
    parser.description = 'Show a holistic access pattern report'
